import	os
import	unittest
from	typing								import Callable
from	random								import randint
from	pygwarts.tests.magical				import MagicalTestCase
from	pygwarts.magical.philosophers_stone	import Transmutable
from	pygwarts.magical.chests				import Chest
from	pygwarts.irma.contrib				import LibraryContrib








class PackingCase(MagicalTestCase):

	"""
		Test cases for magical.chests.Chest
	"""

	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:
			if	os.path.isfile(cls.SET_HANDLER): os.remove(cls.SET_HANDLER)

	@classmethod
	def setUpClass(cls):
		class Treasure(Transmutable):

			weight	:int
			__call__:Callable[[], str]

		class CurrentChest(Chest):
			class loggy(LibraryContrib):

				init_name	= "chest"
				init_level	= 10
				handler		= cls.SET_HANDLER

			class Diamond(Treasure):

				weight	= 100
				def __call__(self): return "is forever?"

			class Ruby(Treasure):

				weight	= 150
				def __call__(self): return "shining red"

			class Jade(Treasure):

				weight	= 200
				def __call__(self): return "ZYO"

			class Sapphire(Treasure):

				weight	= 174
				def __call__(self): return "expense"

			class Gold(Treasure):

				weight	= 995
				def __call__(self): return "The Best"

			class Silver(Treasure):

				weight	= 100500
				def __call__(self): return "never too much"

			class Boot(Transmutable):

				weight	= 500
				def __call__(self): return "was caught in a river, so might be dear to heart"


		cls.make_loggy_file(cls, cls.SET_HANDLER)
		cls.test_case = CurrentChest()
		cls.test_case(cls.test_case.Diamond)
		cls.test_case(cls.test_case.Ruby)
		cls.test_case(cls.test_case.Jade)
		cls.test_case(cls.test_case.Sapphire)
		cls.test_case(cls.test_case.Gold)
		cls.test_case(cls.test_case.Silver)

		class SameChest(Chest):		pass
		class NotSameChest(Chest):
			class Diamond(Treasure):

				weight	= 1002
				def __call__(self): return "same is forever?"

			class Ruby(Treasure):

				weight	= 1502
				def __call__(self): return "same shining red"

			class Jade(Treasure):

				weight	= 2002
				def __call__(self): return "same ZYO"

			class Sapphire(Treasure):

				weight	= 1742
				def __call__(self): return "same expense"

			class Gold(Treasure):

				weight	= 9952
				def __call__(self): return "same The Best"

			class Silver(Treasure):

				weight	= 1005002
				def __call__(self): return "same never too much"

			class Boot(Transmutable):

				weight	= 5002
				def __call__(self): return "same was caught in a river, so might be dear to heart"

		cls.not_same_case = NotSameChest()
		cls.not_same_case(cls.not_same_case.Diamond)
		cls.not_same_case(cls.not_same_case.Ruby)
		cls.not_same_case(cls.not_same_case.Jade)
		cls.not_same_case(cls.not_same_case.Sapphire)
		cls.not_same_case(cls.not_same_case.Gold)
		cls.not_same_case(cls.not_same_case.Silver)

		cls.same_case = SameChest()
		cls.same_case(cls.test_case.Diamond)
		cls.same_case(cls.test_case.Ruby)
		cls.same_case(cls.test_case.Jade)
		cls.same_case(cls.test_case.Sapphire)
		cls.same_case(cls.test_case.Gold)
		cls.same_case(cls.test_case.Silver)

		cls.same_aslist = [

			cls.test_case.Diamond,
			cls.test_case.Ruby,
			cls.test_case.Jade,
			cls.test_case.Sapphire,
			cls.test_case.Gold,
			cls.test_case.Silver
		]
		cls.not_same_aslist = [

			cls.not_same_case.Diamond,
			cls.not_same_case.Ruby,
			cls.not_same_case.Jade,
			cls.not_same_case.Sapphire,
			cls.not_same_case.Gold,
			cls.not_same_case.Silver
		]
		cls.astuple = (

			cls.test_case.Diamond,
			cls.test_case.Ruby,
			cls.test_case.Jade,
			cls.test_case.Sapphire,
			cls.test_case.Gold,
			cls.test_case.Silver
		)




	def test_length(self): self.assertEqual(len(self.test_case), 6)
	def test_inners(self):

		self.assertTrue(hasattr(self.test_case, "_inside_"))
		self.assertIsInstance(self.test_case._inside_, list)
		self.assertTrue(hasattr(self.test_case, "_current"))
		self.assertIsInstance(self.test_case._current, int)




	def test_contains(self):

		self.assertIn(self.test_case.Diamond,	self.test_case)
		self.assertIn(self.test_case.Ruby,		self.test_case)
		self.assertIn(self.test_case.Jade,		self.test_case)
		self.assertIn(self.test_case.Sapphire,	self.test_case)
		self.assertIn(self.test_case.Gold,		self.test_case)
		self.assertIn(self.test_case.Silver,	self.test_case)
		self.assertNotIn(self.test_case.Boot,	self.test_case)




	def test_getitem(self):

		self.assertEqual(self.test_case[0], self.test_case.Diamond)
		self.assertEqual(self.test_case[1], self.test_case.Ruby)
		self.assertEqual(self.test_case[2], self.test_case.Jade)
		self.assertEqual(self.test_case[3], self.test_case.Sapphire)
		self.assertEqual(self.test_case[4], self.test_case.Gold)
		self.assertEqual(self.test_case[5], self.test_case.Silver)
		self.assertEqual(self.test_case[6], None)

		self.assertEqual(self.test_case[-1], self.test_case.Silver)
		self.assertEqual(self.test_case[-2], self.test_case.Gold)
		self.assertEqual(self.test_case[-3], self.test_case.Sapphire)
		self.assertEqual(self.test_case[-4], self.test_case.Jade)
		self.assertEqual(self.test_case[-5], self.test_case.Ruby)
		self.assertEqual(self.test_case[-6], self.test_case.Diamond)
		self.assertEqual(self.test_case[-7], None)




	def test_setitem(self):

		self.assertNotIn(self.test_case.Boot, self.test_case)
		self.assertIn(self.test_case.Sapphire, self.test_case)
		self.assertEqual(self.test_case[3], self.test_case.Sapphire)

		self.test_case[3] = self.test_case.Boot

		self.assertIn(self.test_case.Boot, self.test_case)
		self.assertNotIn(self.test_case.Sapphire, self.test_case)
		self.assertEqual(self.test_case[3], self.test_case.Boot)

		with self.assertLogs("chest", 30) as case_loggy : self.test_case[3] = 42

		self.assertIn(f"WARNING:chest:Can't put \"{type(42)}\" in {self.test_case} Chest", case_loggy.output)
		self.assertIn(self.test_case.Boot, self.test_case)
		self.assertNotIn(self.test_case.Sapphire, self.test_case)
		self.assertEqual(self.test_case[3], self.test_case.Boot)

		self.test_case[3] = self.test_case.Sapphire

		self.assertNotIn(self.test_case.Boot, self.test_case)
		self.assertIn(self.test_case.Sapphire, self.test_case)
		self.assertEqual(self.test_case[3], self.test_case.Sapphire)




	def test_loading_and_deleting(self):

		self.test_case(self.test_case.Boot)
		self.assertEqual(len(self.test_case), 7)
		self.assertEqual(self.test_case[0], self.test_case.Diamond)
		self.assertEqual(self.test_case[1], self.test_case.Ruby)
		self.assertEqual(self.test_case[2], self.test_case.Jade)
		self.assertEqual(self.test_case[3], self.test_case.Sapphire)
		self.assertEqual(self.test_case[4], self.test_case.Gold)
		self.assertEqual(self.test_case[5], self.test_case.Silver)
		self.assertEqual(self.test_case[6], self.test_case.Boot)

		del self.test_case[6]
		self.assertNotIn(self.test_case.Boot, self.test_case)
		self.assertEqual(len(self.test_case), 6)




	def test_unload_and_multiple_load(self):

		self.assertEqual(len(self.test_case), 6)
		self.assertEqual(self.test_case[0], self.test_case.Diamond)
		self.assertEqual(self.test_case[1], self.test_case.Ruby)
		self.assertEqual(self.test_case[2], self.test_case.Jade)
		self.assertEqual(self.test_case[3], self.test_case.Sapphire)
		self.assertEqual(self.test_case[4], self.test_case.Gold)
		self.assertEqual(self.test_case[5], self.test_case.Silver)

		self.test_case.unload()
		self.assertEqual(len(self.test_case), 0)
		self.assertNotIn(self.test_case.Diamond,	self.test_case)
		self.assertNotIn(self.test_case.Ruby,		self.test_case)
		self.assertNotIn(self.test_case.Jade,		self.test_case)
		self.assertNotIn(self.test_case.Sapphire,	self.test_case)
		self.assertNotIn(self.test_case.Gold,		self.test_case)
		self.assertNotIn(self.test_case.Silver,		self.test_case)
		self.assertNotIn(self.test_case.Boot,		self.test_case)

		with self.assertLogs("chest", 10) as case_loggy :
			self.test_case(

				self.test_case.Diamond,
				self.test_case.Ruby,
				self.test_case.Jade,
				69,
				self.test_case.Sapphire,
				self.test_case.Gold,
				self.test_case.Silver
			)

		self.assertEqual(len(self.test_case), 6)
		self.assertEqual(self.test_case[0], self.test_case.Diamond)
		self.assertEqual(self.test_case[1], self.test_case.Ruby)
		self.assertEqual(self.test_case[2], self.test_case.Jade)
		self.assertEqual(self.test_case[3], self.test_case.Sapphire)
		self.assertEqual(self.test_case[4], self.test_case.Gold)
		self.assertEqual(self.test_case[5], self.test_case.Silver)
		self.assertIn(f"DEBUG:chest:Discarding \"{69}\" load", case_loggy.output)

		with self.assertLogs("chest", 30) as case_loggy : self.test_case(42)
		self.assertIn(f"WARNING:chest:Can't put \"{42}\" in {self.test_case} Chest", case_loggy.output)




	def test_index(self):

		self.assertEqual(self.test_case.index(self.test_case.Diamond),	0)
		self.assertEqual(self.test_case.index(self.test_case.Ruby),		1)
		self.assertEqual(self.test_case.index(self.test_case.Jade),		2)
		self.assertEqual(self.test_case.index(self.test_case.Sapphire),	3)
		self.assertEqual(self.test_case.index(self.test_case.Gold),		4)
		self.assertEqual(self.test_case.index(self.test_case.Silver),	5)




	def test_eq_Chest(self):

		self.assertNotEqual(self.test_case, self.not_same_case)
		self.assertEqual(len(self.test_case), len(self.not_same_case))

		self.assertNotEqual(self.test_case[0], self.not_same_case[0])
		self.assertNotEqual(self.test_case[1], self.not_same_case[1])
		self.assertNotEqual(self.test_case[2], self.not_same_case[2])
		self.assertNotEqual(self.test_case[3], self.not_same_case[3])
		self.assertNotEqual(self.test_case[4], self.not_same_case[4])
		self.assertNotEqual(self.test_case[5], self.not_same_case[5])

		self.assertNotEqual(self.test_case[0].weight, self.not_same_case[0].weight)
		self.assertNotEqual(self.test_case[1].weight, self.not_same_case[1].weight)
		self.assertNotEqual(self.test_case[2].weight, self.not_same_case[2].weight)
		self.assertNotEqual(self.test_case[3].weight, self.not_same_case[3].weight)
		self.assertNotEqual(self.test_case[4].weight, self.not_same_case[4].weight)
		self.assertNotEqual(self.test_case[5].weight, self.not_same_case[5].weight)

		self.assertNotEqual(self.test_case[0](), self.not_same_case[0]())
		self.assertNotEqual(self.test_case[1](), self.not_same_case[1]())
		self.assertNotEqual(self.test_case[2](), self.not_same_case[2]())
		self.assertNotEqual(self.test_case[3](), self.not_same_case[3]())
		self.assertNotEqual(self.test_case[4](), self.not_same_case[4]())
		self.assertNotEqual(self.test_case[5](), self.not_same_case[5]())


		self.assertEqual(self.test_case, self.same_case)
		self.assertEqual(len(self.test_case), len(self.same_case))

		self.assertEqual(self.test_case[0], self.same_case[0])
		self.assertEqual(self.test_case[1], self.same_case[1])
		self.assertEqual(self.test_case[2], self.same_case[2])
		self.assertEqual(self.test_case[3], self.same_case[3])
		self.assertEqual(self.test_case[4], self.same_case[4])
		self.assertEqual(self.test_case[5], self.same_case[5])

		self.assertEqual(self.test_case[0].weight, self.same_case[0].weight)
		self.assertEqual(self.test_case[1].weight, self.same_case[1].weight)
		self.assertEqual(self.test_case[2].weight, self.same_case[2].weight)
		self.assertEqual(self.test_case[3].weight, self.same_case[3].weight)
		self.assertEqual(self.test_case[4].weight, self.same_case[4].weight)
		self.assertEqual(self.test_case[5].weight, self.same_case[5].weight)

		self.assertEqual(self.test_case[0](), self.same_case[0]())
		self.assertEqual(self.test_case[1](), self.same_case[1]())
		self.assertEqual(self.test_case[2](), self.same_case[2]())
		self.assertEqual(self.test_case[3](), self.same_case[3]())
		self.assertEqual(self.test_case[4](), self.same_case[4]())
		self.assertEqual(self.test_case[5](), self.same_case[5]())




	def test_gt_ge_lt_le_Chest(self):

		self.assertEqual(self.not_same_case[0], self.not_same_case.Diamond)
		self.assertEqual(self.not_same_case[1], self.not_same_case.Ruby)
		self.assertEqual(self.not_same_case[2], self.not_same_case.Jade)
		self.assertEqual(self.not_same_case[3], self.not_same_case.Sapphire)
		self.assertEqual(self.not_same_case[4], self.not_same_case.Gold)
		self.assertEqual(self.not_same_case[5], self.not_same_case.Silver)
		del self.not_same_case[0]

		self.assertNotEqual(self.test_case, self.not_same_case)
		self.assertNotEqual(len(self.test_case), len(self.not_same_case))
		self.assertGreater(self.test_case, self.not_same_case)
		self.assertGreaterEqual(self.test_case, self.not_same_case)
		self.assertLess(self.not_same_case, self.test_case)
		self.assertLessEqual(self.not_same_case, self.test_case)

		while len(self.not_same_case): del self.not_same_case[0]
		self.not_same_case(self.not_same_case.Diamond)
		self.not_same_case(self.not_same_case.Ruby)
		self.not_same_case(self.not_same_case.Jade)
		self.not_same_case(self.not_same_case.Sapphire)
		self.not_same_case(self.not_same_case.Gold)
		self.not_same_case(self.not_same_case.Silver)

		self.assertEqual(self.not_same_case[0], self.not_same_case.Diamond)
		self.assertEqual(self.not_same_case[1], self.not_same_case.Ruby)
		self.assertEqual(self.not_same_case[2], self.not_same_case.Jade)
		self.assertEqual(self.not_same_case[3], self.not_same_case.Sapphire)
		self.assertEqual(self.not_same_case[4], self.not_same_case.Gold)
		self.assertEqual(self.not_same_case[5], self.not_same_case.Silver)




	def test_eq_list(self):

		self.assertNotEqual(self.test_case, self.not_same_aslist)
		self.assertEqual(len(self.test_case), len(self.not_same_aslist))

		self.assertNotEqual(self.test_case[0], self.not_same_aslist[0])
		self.assertNotEqual(self.test_case[1], self.not_same_aslist[1])
		self.assertNotEqual(self.test_case[2], self.not_same_aslist[2])
		self.assertNotEqual(self.test_case[3], self.not_same_aslist[3])
		self.assertNotEqual(self.test_case[4], self.not_same_aslist[4])
		self.assertNotEqual(self.test_case[5], self.not_same_aslist[5])

		self.assertNotEqual(self.test_case[0].weight, self.not_same_aslist[0].weight)
		self.assertNotEqual(self.test_case[1].weight, self.not_same_aslist[1].weight)
		self.assertNotEqual(self.test_case[2].weight, self.not_same_aslist[2].weight)
		self.assertNotEqual(self.test_case[3].weight, self.not_same_aslist[3].weight)
		self.assertNotEqual(self.test_case[4].weight, self.not_same_aslist[4].weight)
		self.assertNotEqual(self.test_case[5].weight, self.not_same_aslist[5].weight)

		self.assertNotEqual(self.test_case[0](), self.not_same_aslist[0]())
		self.assertNotEqual(self.test_case[1](), self.not_same_aslist[1]())
		self.assertNotEqual(self.test_case[2](), self.not_same_aslist[2]())
		self.assertNotEqual(self.test_case[3](), self.not_same_aslist[3]())
		self.assertNotEqual(self.test_case[4](), self.not_same_aslist[4]())
		self.assertNotEqual(self.test_case[5](), self.not_same_aslist[5]())


		self.assertEqual(self.test_case, self.same_aslist)
		self.assertEqual(len(self.test_case), len(self.same_aslist))

		self.assertEqual(self.test_case[0], self.same_aslist[0])
		self.assertEqual(self.test_case[1], self.same_aslist[1])
		self.assertEqual(self.test_case[2], self.same_aslist[2])
		self.assertEqual(self.test_case[3], self.same_aslist[3])
		self.assertEqual(self.test_case[4], self.same_aslist[4])
		self.assertEqual(self.test_case[5], self.same_aslist[5])

		self.assertEqual(self.test_case[0].weight, self.same_aslist[0].weight)
		self.assertEqual(self.test_case[1].weight, self.same_aslist[1].weight)
		self.assertEqual(self.test_case[2].weight, self.same_aslist[2].weight)
		self.assertEqual(self.test_case[3].weight, self.same_aslist[3].weight)
		self.assertEqual(self.test_case[4].weight, self.same_aslist[4].weight)
		self.assertEqual(self.test_case[5].weight, self.same_aslist[5].weight)

		self.assertEqual(self.test_case[0](), self.same_aslist[0]())
		self.assertEqual(self.test_case[1](), self.same_aslist[1]())
		self.assertEqual(self.test_case[2](), self.same_aslist[2]())
		self.assertEqual(self.test_case[3](), self.same_aslist[3]())
		self.assertEqual(self.test_case[4](), self.same_aslist[4]())
		self.assertEqual(self.test_case[5](), self.same_aslist[5]())




	def test_gt_ge_lt_le_list(self):

		self.assertEqual(self.not_same_aslist[0], self.not_same_case.Diamond)
		self.assertEqual(self.not_same_aslist[1], self.not_same_case.Ruby)
		self.assertEqual(self.not_same_aslist[2], self.not_same_case.Jade)
		self.assertEqual(self.not_same_aslist[3], self.not_same_case.Sapphire)
		self.assertEqual(self.not_same_aslist[4], self.not_same_case.Gold)
		self.assertEqual(self.not_same_aslist[5], self.not_same_case.Silver)
		del self.not_same_aslist[0]

		self.assertNotEqual(self.test_case, self.not_same_aslist)
		self.assertNotEqual(len(self.test_case), len(self.not_same_aslist))
		self.assertGreater(self.test_case, self.not_same_aslist)
		self.assertGreaterEqual(self.test_case, self.not_same_aslist)
		self.assertLess(self.not_same_aslist, self.test_case)
		self.assertLessEqual(self.not_same_aslist, self.test_case)

		self.not_same_aslist.insert(0, self.not_same_case.Diamond)
		self.assertEqual(self.not_same_aslist[0], self.not_same_case.Diamond)
		self.assertEqual(self.not_same_aslist[1], self.not_same_case.Ruby)
		self.assertEqual(self.not_same_aslist[2], self.not_same_case.Jade)
		self.assertEqual(self.not_same_aslist[3], self.not_same_case.Sapphire)
		self.assertEqual(self.not_same_aslist[4], self.not_same_case.Gold)
		self.assertEqual(self.not_same_aslist[5], self.not_same_case.Silver)




	def test_iter(self):

		count = 0
		for item in self.test_case:

			self.assertIsInstance(item.weight, int)
			self.assertIsInstance(item(), str)
			count += 1

		self.assertEqual(count, len(self.test_case))
		count = 0
		_case = iter(self.test_case)
		for one,two in zip(_case, _case):

			self.assertIsInstance(one.weight, int)
			self.assertIsInstance(one(), str)
			count += 1
			self.assertIsInstance(two.weight, int)
			self.assertIsInstance(two(), str)
			count += 1

		self.assertEqual(count, len(self.test_case))




	def test_next(self):

		count = 0
		for _ in range(3 *len(self.test_case)):

			try:

				next(self.test_case)
				count += 1

			except	StopIteration:

				self.assertEqual(count, len(self.test_case))
				count = 0




	def test_noload_call(self):

		test_case = self.test_case()
		self.assertIsInstance(test_case, list)
		self.assertEqual(test_case, self.test_case)
		self.assertNotEqual(id(test_case), id(self.test_case))
		self.assertEqual(test_case[0], self.test_case[0])
		self.assertEqual(test_case[1], self.test_case[1])
		self.assertEqual(test_case[2], self.test_case[2])
		self.assertEqual(test_case[3], self.test_case[3])
		self.assertEqual(test_case[4], self.test_case[4])
		self.assertEqual(test_case[5], self.test_case[5])
















class SelfPackedCase(MagicalTestCase):

	"""
		Test cases for magical.chests.Chest
	"""

	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:
			if	os.path.isfile(cls.PACKED_SET_HANDLER): os.remove(cls.PACKED_SET_HANDLER)

	@classmethod
	def setUpClass(cls):
		class Treasure(Transmutable):

			weight	:int
			__call__:Callable[[], str]

			def __init__(self, *args, **kwargs):

				super().__init__(*args, **kwargs)
				if	isinstance(self._UPPER_LAYER, Chest): self._UPPER_LAYER(self)

		class CurrentChest(Chest):
			class loggy(LibraryContrib):

				init_name	= "pchest"
				init_level	= 10
				handler		= cls.PACKED_SET_HANDLER

			class Diamond(Treasure):

				weight	= 100
				def __call__(self): return "is forever?"

			class Ruby(Treasure):

				weight	= 150
				def __call__(self): return "shining red"

			class Jade(Treasure):

				weight	= 200
				def __call__(self): return "ZYO"

			class Sapphire(Treasure):

				weight	= 174
				def __call__(self): return "expense"

			class Gold(Treasure):

				weight	= 995
				def __call__(self): return "The Best"

			class Silver(Treasure):

				weight	= 100500
				def __call__(self): return "never too much"

			class Boot(Transmutable):

				weight	= 500
				def __call__(self): return "was caught in a river, so might be dear to heart"


		cls.make_loggy_file(cls, cls.PACKED_SET_HANDLER)
		cls.test_case = CurrentChest()

		class SameChest(Chest):		pass
		class NotSameChest(Chest):
			class Diamond(Treasure):

				weight	= 1002
				def __call__(self): return "same is forever?"

			class Ruby(Treasure):

				weight	= 1502
				def __call__(self): return "same shining red"

			class Jade(Treasure):

				weight	= 2002
				def __call__(self): return "same ZYO"

			class Sapphire(Treasure):

				weight	= 1742
				def __call__(self): return "same expense"

			class Gold(Treasure):

				weight	= 9952
				def __call__(self): return "same The Best"

			class Silver(Treasure):

				weight	= 1005002
				def __call__(self): return "same never too much"

			class Boot(Transmutable):

				weight	= 5002
				def __call__(self): return "same was caught in a river, so might be dear to heart"

		cls.not_same_case = NotSameChest()
		cls.same_case = SameChest()
		cls.same_case(cls.test_case.Diamond)
		cls.same_case(cls.test_case.Ruby)
		cls.same_case(cls.test_case.Jade)
		cls.same_case(cls.test_case.Sapphire)
		cls.same_case(cls.test_case.Gold)
		cls.same_case(cls.test_case.Silver)

		cls.same_aslist = [

			cls.test_case.Diamond,
			cls.test_case.Ruby,
			cls.test_case.Jade,
			cls.test_case.Sapphire,
			cls.test_case.Gold,
			cls.test_case.Silver
		]
		cls.not_same_aslist = [

			cls.not_same_case.Diamond,
			cls.not_same_case.Ruby,
			cls.not_same_case.Jade,
			cls.not_same_case.Sapphire,
			cls.not_same_case.Gold,
			cls.not_same_case.Silver
		]
		cls.astuple = (

			cls.test_case.Diamond,
			cls.test_case.Ruby,
			cls.test_case.Jade,
			cls.test_case.Sapphire,
			cls.test_case.Gold,
			cls.test_case.Silver
		)




	def test_packed_length(self): self.assertEqual(len(self.test_case), 6)
	def test_inners(self):

		self.assertTrue(hasattr(self.test_case, "_inside_"))
		self.assertIsInstance(self.test_case._inside_, list)
		self.assertTrue(hasattr(self.test_case, "_current"))
		self.assertIsInstance(self.test_case._current, int)




	def test_packed_contains(self):

		self.assertIn(self.test_case.Diamond,	self.test_case)
		self.assertIn(self.test_case.Ruby,		self.test_case)
		self.assertIn(self.test_case.Jade,		self.test_case)
		self.assertIn(self.test_case.Sapphire,	self.test_case)
		self.assertIn(self.test_case.Gold,		self.test_case)
		self.assertIn(self.test_case.Silver,	self.test_case)
		self.assertNotIn(self.test_case.Boot,	self.test_case)




	def test_packed_getitem(self):

		self.assertEqual(self.test_case[0], self.test_case.Diamond)
		self.assertEqual(self.test_case[1], self.test_case.Ruby)
		self.assertEqual(self.test_case[2], self.test_case.Jade)
		self.assertEqual(self.test_case[3], self.test_case.Sapphire)
		self.assertEqual(self.test_case[4], self.test_case.Gold)
		self.assertEqual(self.test_case[5], self.test_case.Silver)
		self.assertEqual(self.test_case[6], None)

		self.assertEqual(self.test_case[-1], self.test_case.Silver)
		self.assertEqual(self.test_case[-2], self.test_case.Gold)
		self.assertEqual(self.test_case[-3], self.test_case.Sapphire)
		self.assertEqual(self.test_case[-4], self.test_case.Jade)
		self.assertEqual(self.test_case[-5], self.test_case.Ruby)
		self.assertEqual(self.test_case[-6], self.test_case.Diamond)
		self.assertEqual(self.test_case[-7], None)




	def test_packed_setitem(self):

		self.assertNotIn(self.test_case.Boot, self.test_case)
		self.assertIn(self.test_case.Sapphire, self.test_case)
		self.assertEqual(self.test_case[3], self.test_case.Sapphire)

		self.test_case[3] = self.test_case.Boot

		self.assertIn(self.test_case.Boot, self.test_case)
		self.assertNotIn(self.test_case.Sapphire, self.test_case)
		self.assertEqual(self.test_case[3], self.test_case.Boot)

		with self.assertLogs("pchest", 30) as case_loggy : self.test_case[3] = 42

		self.assertIn(f"WARNING:pchest:Can't put \"{type(42)}\" in {self.test_case} Chest", case_loggy.output)
		self.assertIn(self.test_case.Boot, self.test_case)
		self.assertNotIn(self.test_case.Sapphire, self.test_case)
		self.assertEqual(self.test_case[3], self.test_case.Boot)

		self.test_case[3] = self.test_case.Sapphire

		self.assertNotIn(self.test_case.Boot, self.test_case)
		self.assertIn(self.test_case.Sapphire, self.test_case)
		self.assertEqual(self.test_case[3], self.test_case.Sapphire)




	def test_packed_loading_and_deleting(self):

		self.test_case(self.test_case.Boot)
		self.assertEqual(len(self.test_case), 7)
		self.assertEqual(self.test_case[0], self.test_case.Diamond)
		self.assertEqual(self.test_case[1], self.test_case.Ruby)
		self.assertEqual(self.test_case[2], self.test_case.Jade)
		self.assertEqual(self.test_case[3], self.test_case.Sapphire)
		self.assertEqual(self.test_case[4], self.test_case.Gold)
		self.assertEqual(self.test_case[5], self.test_case.Silver)
		self.assertEqual(self.test_case[6], self.test_case.Boot)

		del self.test_case[6]
		self.assertNotIn(self.test_case.Boot, self.test_case)
		self.assertEqual(len(self.test_case), 6)




	def test_packed_unload_and_multiple_load(self):

		self.assertEqual(len(self.test_case), 6)
		self.assertEqual(self.test_case[0], self.test_case.Diamond)
		self.assertEqual(self.test_case[1], self.test_case.Ruby)
		self.assertEqual(self.test_case[2], self.test_case.Jade)
		self.assertEqual(self.test_case[3], self.test_case.Sapphire)
		self.assertEqual(self.test_case[4], self.test_case.Gold)
		self.assertEqual(self.test_case[5], self.test_case.Silver)

		self.test_case.unload()
		self.assertEqual(len(self.test_case), 0)
		self.assertNotIn(self.test_case.Diamond,	self.test_case)
		self.assertNotIn(self.test_case.Ruby,		self.test_case)
		self.assertNotIn(self.test_case.Jade,		self.test_case)
		self.assertNotIn(self.test_case.Sapphire,	self.test_case)
		self.assertNotIn(self.test_case.Gold,		self.test_case)
		self.assertNotIn(self.test_case.Silver,		self.test_case)
		self.assertNotIn(self.test_case.Boot,		self.test_case)

		with self.assertLogs("pchest", 10) as case_loggy :
			self.test_case(

				self.test_case.Diamond,
				self.test_case.Ruby,
				self.test_case.Jade,
				69,
				self.test_case.Sapphire,
				self.test_case.Gold,
				self.test_case.Silver
			)

		self.assertEqual(len(self.test_case), 6)
		self.assertEqual(self.test_case[0], self.test_case.Diamond)
		self.assertEqual(self.test_case[1], self.test_case.Ruby)
		self.assertEqual(self.test_case[2], self.test_case.Jade)
		self.assertEqual(self.test_case[3], self.test_case.Sapphire)
		self.assertEqual(self.test_case[4], self.test_case.Gold)
		self.assertEqual(self.test_case[5], self.test_case.Silver)
		self.assertIn(f"DEBUG:pchest:Discarding \"{69}\" load", case_loggy.output)

		with self.assertLogs("pchest", 30) as case_loggy : self.test_case(42)
		self.assertIn(f"WARNING:pchest:Can't put \"{42}\" in {self.test_case} Chest", case_loggy.output)




	def test_packed_index(self):

		self.assertEqual(self.test_case.index(self.test_case.Diamond),	0)
		self.assertEqual(self.test_case.index(self.test_case.Ruby),		1)
		self.assertEqual(self.test_case.index(self.test_case.Jade),		2)
		self.assertEqual(self.test_case.index(self.test_case.Sapphire),	3)
		self.assertEqual(self.test_case.index(self.test_case.Gold),		4)
		self.assertEqual(self.test_case.index(self.test_case.Silver),	5)




	def test_packed_eq_Chest(self):

		self.assertNotEqual(self.test_case, self.not_same_case)
		self.assertEqual(len(self.test_case), len(self.not_same_case))

		self.assertNotEqual(self.test_case[0], self.not_same_case[0])
		self.assertNotEqual(self.test_case[1], self.not_same_case[1])
		self.assertNotEqual(self.test_case[2], self.not_same_case[2])
		self.assertNotEqual(self.test_case[3], self.not_same_case[3])
		self.assertNotEqual(self.test_case[4], self.not_same_case[4])
		self.assertNotEqual(self.test_case[5], self.not_same_case[5])

		self.assertNotEqual(self.test_case[0].weight, self.not_same_case[0].weight)
		self.assertNotEqual(self.test_case[1].weight, self.not_same_case[1].weight)
		self.assertNotEqual(self.test_case[2].weight, self.not_same_case[2].weight)
		self.assertNotEqual(self.test_case[3].weight, self.not_same_case[3].weight)
		self.assertNotEqual(self.test_case[4].weight, self.not_same_case[4].weight)
		self.assertNotEqual(self.test_case[5].weight, self.not_same_case[5].weight)

		self.assertNotEqual(self.test_case[0](), self.not_same_case[0]())
		self.assertNotEqual(self.test_case[1](), self.not_same_case[1]())
		self.assertNotEqual(self.test_case[2](), self.not_same_case[2]())
		self.assertNotEqual(self.test_case[3](), self.not_same_case[3]())
		self.assertNotEqual(self.test_case[4](), self.not_same_case[4]())
		self.assertNotEqual(self.test_case[5](), self.not_same_case[5]())


		self.assertEqual(self.test_case, self.same_case)
		self.assertEqual(len(self.test_case), len(self.same_case))

		self.assertEqual(self.test_case[0], self.same_case[0])
		self.assertEqual(self.test_case[1], self.same_case[1])
		self.assertEqual(self.test_case[2], self.same_case[2])
		self.assertEqual(self.test_case[3], self.same_case[3])
		self.assertEqual(self.test_case[4], self.same_case[4])
		self.assertEqual(self.test_case[5], self.same_case[5])

		self.assertEqual(self.test_case[0].weight, self.same_case[0].weight)
		self.assertEqual(self.test_case[1].weight, self.same_case[1].weight)
		self.assertEqual(self.test_case[2].weight, self.same_case[2].weight)
		self.assertEqual(self.test_case[3].weight, self.same_case[3].weight)
		self.assertEqual(self.test_case[4].weight, self.same_case[4].weight)
		self.assertEqual(self.test_case[5].weight, self.same_case[5].weight)

		self.assertEqual(self.test_case[0](), self.same_case[0]())
		self.assertEqual(self.test_case[1](), self.same_case[1]())
		self.assertEqual(self.test_case[2](), self.same_case[2]())
		self.assertEqual(self.test_case[3](), self.same_case[3]())
		self.assertEqual(self.test_case[4](), self.same_case[4]())
		self.assertEqual(self.test_case[5](), self.same_case[5]())




	def test_packed_gt_ge_lt_le_Chest(self):

		self.assertEqual(self.not_same_case[0], self.not_same_case.Diamond)
		self.assertEqual(self.not_same_case[1], self.not_same_case.Ruby)
		self.assertEqual(self.not_same_case[2], self.not_same_case.Jade)
		self.assertEqual(self.not_same_case[3], self.not_same_case.Sapphire)
		self.assertEqual(self.not_same_case[4], self.not_same_case.Gold)
		self.assertEqual(self.not_same_case[5], self.not_same_case.Silver)
		del self.not_same_case[0]

		self.assertNotEqual(self.test_case, self.not_same_case)
		self.assertNotEqual(len(self.test_case), len(self.not_same_case))
		self.assertGreater(self.test_case, self.not_same_case)
		self.assertGreaterEqual(self.test_case, self.not_same_case)
		self.assertLess(self.not_same_case, self.test_case)
		self.assertLessEqual(self.not_same_case, self.test_case)

		while len(self.not_same_case): del self.not_same_case[0]
		self.not_same_case(self.not_same_case.Diamond)
		self.not_same_case(self.not_same_case.Ruby)
		self.not_same_case(self.not_same_case.Jade)
		self.not_same_case(self.not_same_case.Sapphire)
		self.not_same_case(self.not_same_case.Gold)
		self.not_same_case(self.not_same_case.Silver)

		self.assertEqual(self.not_same_case[0], self.not_same_case.Diamond)
		self.assertEqual(self.not_same_case[1], self.not_same_case.Ruby)
		self.assertEqual(self.not_same_case[2], self.not_same_case.Jade)
		self.assertEqual(self.not_same_case[3], self.not_same_case.Sapphire)
		self.assertEqual(self.not_same_case[4], self.not_same_case.Gold)
		self.assertEqual(self.not_same_case[5], self.not_same_case.Silver)




	def test_packed_eq_list(self):

		self.assertNotEqual(self.test_case, self.not_same_aslist)
		self.assertEqual(len(self.test_case), len(self.not_same_aslist))

		self.assertNotEqual(self.test_case[0], self.not_same_aslist[0])
		self.assertNotEqual(self.test_case[1], self.not_same_aslist[1])
		self.assertNotEqual(self.test_case[2], self.not_same_aslist[2])
		self.assertNotEqual(self.test_case[3], self.not_same_aslist[3])
		self.assertNotEqual(self.test_case[4], self.not_same_aslist[4])
		self.assertNotEqual(self.test_case[5], self.not_same_aslist[5])

		self.assertNotEqual(self.test_case[0].weight, self.not_same_aslist[0].weight)
		self.assertNotEqual(self.test_case[1].weight, self.not_same_aslist[1].weight)
		self.assertNotEqual(self.test_case[2].weight, self.not_same_aslist[2].weight)
		self.assertNotEqual(self.test_case[3].weight, self.not_same_aslist[3].weight)
		self.assertNotEqual(self.test_case[4].weight, self.not_same_aslist[4].weight)
		self.assertNotEqual(self.test_case[5].weight, self.not_same_aslist[5].weight)

		self.assertNotEqual(self.test_case[0](), self.not_same_aslist[0]())
		self.assertNotEqual(self.test_case[1](), self.not_same_aslist[1]())
		self.assertNotEqual(self.test_case[2](), self.not_same_aslist[2]())
		self.assertNotEqual(self.test_case[3](), self.not_same_aslist[3]())
		self.assertNotEqual(self.test_case[4](), self.not_same_aslist[4]())
		self.assertNotEqual(self.test_case[5](), self.not_same_aslist[5]())


		self.assertEqual(self.test_case, self.same_aslist)
		self.assertEqual(len(self.test_case), len(self.same_aslist))

		self.assertEqual(self.test_case[0], self.same_aslist[0])
		self.assertEqual(self.test_case[1], self.same_aslist[1])
		self.assertEqual(self.test_case[2], self.same_aslist[2])
		self.assertEqual(self.test_case[3], self.same_aslist[3])
		self.assertEqual(self.test_case[4], self.same_aslist[4])
		self.assertEqual(self.test_case[5], self.same_aslist[5])

		self.assertEqual(self.test_case[0].weight, self.same_aslist[0].weight)
		self.assertEqual(self.test_case[1].weight, self.same_aslist[1].weight)
		self.assertEqual(self.test_case[2].weight, self.same_aslist[2].weight)
		self.assertEqual(self.test_case[3].weight, self.same_aslist[3].weight)
		self.assertEqual(self.test_case[4].weight, self.same_aslist[4].weight)
		self.assertEqual(self.test_case[5].weight, self.same_aslist[5].weight)

		self.assertEqual(self.test_case[0](), self.same_aslist[0]())
		self.assertEqual(self.test_case[1](), self.same_aslist[1]())
		self.assertEqual(self.test_case[2](), self.same_aslist[2]())
		self.assertEqual(self.test_case[3](), self.same_aslist[3]())
		self.assertEqual(self.test_case[4](), self.same_aslist[4]())
		self.assertEqual(self.test_case[5](), self.same_aslist[5]())




	def test_packed_gt_ge_lt_le_list(self):

		self.assertEqual(self.not_same_aslist[0], self.not_same_case.Diamond)
		self.assertEqual(self.not_same_aslist[1], self.not_same_case.Ruby)
		self.assertEqual(self.not_same_aslist[2], self.not_same_case.Jade)
		self.assertEqual(self.not_same_aslist[3], self.not_same_case.Sapphire)
		self.assertEqual(self.not_same_aslist[4], self.not_same_case.Gold)
		self.assertEqual(self.not_same_aslist[5], self.not_same_case.Silver)
		del self.not_same_aslist[0]

		self.assertNotEqual(self.test_case, self.not_same_aslist)
		self.assertNotEqual(len(self.test_case), len(self.not_same_aslist))
		self.assertGreater(self.test_case, self.not_same_aslist)
		self.assertGreaterEqual(self.test_case, self.not_same_aslist)
		self.assertLess(self.not_same_aslist, self.test_case)
		self.assertLessEqual(self.not_same_aslist, self.test_case)

		self.not_same_aslist.insert(0, self.not_same_case.Diamond)
		self.assertEqual(self.not_same_aslist[0], self.not_same_case.Diamond)
		self.assertEqual(self.not_same_aslist[1], self.not_same_case.Ruby)
		self.assertEqual(self.not_same_aslist[2], self.not_same_case.Jade)
		self.assertEqual(self.not_same_aslist[3], self.not_same_case.Sapphire)
		self.assertEqual(self.not_same_aslist[4], self.not_same_case.Gold)
		self.assertEqual(self.not_same_aslist[5], self.not_same_case.Silver)




	def test_iter(self):

		count = 0
		for item in self.test_case:

			self.assertIsInstance(item.weight, int)
			self.assertIsInstance(item(), str)
			count += 1

		self.assertEqual(count, len(self.test_case))
		count = 0
		_case = iter(self.test_case)
		for one,two in zip(_case, _case):

			self.assertIsInstance(one.weight, int)
			self.assertIsInstance(one(), str)
			count += 1
			self.assertIsInstance(two.weight, int)
			self.assertIsInstance(two(), str)
			count += 1

		self.assertEqual(count, len(self.test_case))




	def test_next(self):

		count = 0
		for _ in range(3 *len(self.test_case)):

			try:

				next(self.test_case)
				count += 1

			except	StopIteration:

				self.assertEqual(count, len(self.test_case))
				count = 0




	def test_packed_noload_call(self):

		test_case = self.test_case()
		self.assertIsInstance(test_case, list)
		self.assertEqual(test_case, self.test_case)
		self.assertNotEqual(id(test_case), id(self.test_case))
		self.assertEqual(test_case[0], self.test_case[0])
		self.assertEqual(test_case[1], self.test_case[1])
		self.assertEqual(test_case[2], self.test_case[2])
		self.assertEqual(test_case[3], self.test_case[3])
		self.assertEqual(test_case[4], self.test_case[4])
		self.assertEqual(test_case[5], self.test_case[5])








if	__name__ == "__main__" : unittest.main(verbosity=2)










#                                                                                                              
#                                                                                                              
#                                                                                                              
#                                                                                                              
#                                                      .                                                       
#                                                      }d,                                                     
#                                                      1a'                                                     
#                                                      (k.                                                     
#                                                      rm"                                                     
#                                                     'Xz<.                                                    
#                                                   "n+0Yn{[.                                                  
#                                                 .-Qdn]#+ahc^                                                 
#                                                 \Ckk]-#,QkkmI                                                
#                                               ,xLhZ> )*`'jkkd}.                                              
#                                              ]cOhn'  cd; .~wkhz".                                            
#                                            ^jcbq+   'Uv-.  `uhkq!                                            
#                                           ,XzhQ,    ubqQ>    {kkq;                                           
#                                          ,nLhj'.    >uUx`    .>mkd[                                          
#                                         ]uOm~,~}/tt/!ra/)/tt(-!^Uhhc".                                       
#                                       `uUxvrr/t//nXU~Ch)vYzxfrrrxcuvO_.                                      
#                                     .")/Lu]j0dX/?l` .kd{..^!]fUbX)\Lut<.                                     
#                                    .:rU)(CQ)!.      "oL(       '_uqzjuJ].                                    
#                                  .`rC]xbc<         '/OUv>.        .?Jdjx0~                                   
#                                 .iZ((dUl           :O0JC(.           ~0q\Ox'.                                
#                                 lp]uqf.            .>tLu.             'zbtZu                                 
#                               ",Zj\dn.              :zbn.              'Jk/h[l.                              
#                              _C1b!hq'               ^mkr                ;abuq?z^                             
#                            `xvnzYjku                ;ak|                 Lhv*!ow~                            
#                          .:C/bjXYxkn                }a0r.                CkCo!akh(.                          
#                          ?XXax`td)hO.              iCQLJj`              `kkZk`~wkhz^                         
#                        `xXwq_. "bfwh1              -hOJCn:            ..nkkkf..`vhkq+.                       
#                      .~UChz"    _hxdh{.            .(uOZ`.           ..rhkkJ'    ?pkhx'                      
#                     .(Lqh|.      -hcphn'            [UkL             "JkkkJ'      :Zkku.                     
#                     (Obw+.       .iQOZkm)`          _ZkU          ."/dkkbt'        `nhhc^                    
#                   `nqhc^.         ..]Qqdkdj!..      {kkz       .^{Xbkkbx:.          ._dkq+.                  
#                  <mkd]                ixwkkkpX/]>". xwOt:..,<}fUbkkkU{`               >pkhx'                 
#                'fkdZcffffffffffffffffffffczCbkkkkhj/pqqZC?qhkkkh0czxfffffffffffffffffffnQmbZ>                
#              ."OoaoaaaaaaaaaaaaaaaaaaaaaaaopXccvc0>cqZUZv]nYuczcZoaaaaaaaaaaaaaaaaaaaaaaaoaao{.              
#                                                 .  :LYc|1.  .                                                
#                                                    .*bpc[                                                    
#                                                    'hkkd|                                                    
#                                                     ~hkY .                                                   
#                                                     l/(-.    
#
# Ain't fan!


from typing									import List
from typing									import Dict
from pygwarts.magical.philosophers_stone	import Transmutable
from pygwarts.magical.chests				import KeyChest
from pygwarts.magical.spells				import flagrate








class LetterBuilder(KeyChest):

	"""
		hedwig.mail core class that is a super class for letter composing objects. The KeyChest base
		allows every descent ObjectField letter field to be included for a build. That including takes
		place during initiation and further functionality implemented in following members:
			Field			- the object which upon called on a letter field will produce the final
							value for such letter field; this is the general way for building fields;
			mandatoryField	- additional field value gathering object, that makes use of "Field" and
							extends it with result verification, so means it is used for letter fields
							that must return some value;
			is_buildable	- helper method that validates the consistency of a dictionary that is to
							be obtained from a "build" method;
			build			- core method that initiates the process of letter building by invoking
							every letter field; is a super method for every builder.
	"""

	class Field(Transmutable):

		"""
			hedwig.mail letter fields main obtainer. Accepts a string "field_name" that must represent the
			key to be searched in upper layer KeyChest LetterBuilder. Also accepts "as_string" boolean
			flag that decides whether output value will be concatenated string of a list of strings right
			before concatenation. In other words "Field" gather mapped with "field_name" fields invocations
			results, and either return it as a list or as a joined string.
		"""


		def __call__(self, field_name :str, as_string=True) -> str | List[str] | None :
			self.loggy.debug(f"Obtaining {field_name} field value")


			if	hasattr(self, "_UPPER_LAYER") and isinstance(self._UPPER_LAYER, KeyChest):
				if	isinstance(field_source := self._UPPER_LAYER[field_name], list):
					current_field = str() if as_string else list()


					# Every item in "field_name" mapping must be a callable that will return string.
					# It is the interface for the letter fields objects "Field" relies on. Any other
					# items encountered will be skipped.
					for current_source in field_source:
						if	callable(current_source):
							if	isinstance(current_value := current_source(),str):


								current_field += current_value if as_string else [ current_value ]


							else:	self.loggy.debug(f"{current_source} returned not a string")
						else:		self.loggy.debug(f"{current_source} is not callable")


					flen = len(current_field)
					self.loggy.info(

						"%s field obtained %s %s%s"%(
							field_name,
							flen,
							"symbol" if as_string else "item",
							flagrate(flen)
						)
					)


					# It is a probable situation, that "current_field" end up as an empty string or list,
					# cause every processed field was just skipped, so any obtainer must account for
					# this falsy variant.
					return	current_field
				else:		self.loggy.debug(f"{field_name} field is not a list")
			else:			self.loggy.debug(f"{field_name} field cannot be included for building")




	class mandatoryField(Transmutable):

		"""
			hedwig.mail letter fields additional obtainer. Acts the same way as "Field" cause invokes it,
			but also logs a critical level message if "Field" returns a falsy value, like None it case
			"Field" could not obtain field value, or an empty list or string if all fields was skipped.
		"""

		def __call__(self, field_name :str, as_string=True) -> str | None :

			if		(field_value := self.Field(field_name, as_string)): return field_value
			else:	self.loggy.critical(f"{self._UPPER_LAYER} {field_name} field failed")




	def is_buildable(self, candidate :Dict[str,str | List[str]]) -> bool :

		"""
			Helper method that validates a dictionary that represents current letter build, which is mapping
			of typical letter fields names with corresponding values that "Field" object produces after such
			field names processing. Basically this dictionary must have 7 key-value pairs, 2 of them are
			mandatory and the rest is optional, that means for mandatory values except for optional values,
			None is not allowed. Returns boolean values that represents validation status.
		"""

		return	(

			isinstance(candidate, dict)							and
			len(candidate) == 7									and
			isinstance(candidate.get("sender"), list)			and
			isinstance(candidate.get("recipient"), str)			and
			isinstance(candidate.get("cc"), str | None)			and
			isinstance(candidate.get("bcc"), str | None)		and
			isinstance(candidate.get("subject"), str | None)	and
			isinstance(candidate.get("body"), str | None)		and
			isinstance(candidate.get("attachment"), list | None)
		)




	def build(self) -> Dict[str,str | List[str]] | None :

		"""
			The core method which is super for builders. Gathers all possible regular field names and
			creates a dictionary mapping to be returned. First verifies mandatory fields like "sender"
			and "recipient" are obtained, then goes for optional fields "cc", "bcc", "subject", "body"
			and "attachment". Returns compatible with "is_buildable" method check dictionary in case
			mandatory fields are obtained, None otherwise.
		"""

		if	isinstance(sender := self.mandatoryField("sender", as_string=False), list):
			if	isinstance(recipient := self.mandatoryField("recipient"), str):

				return	{

					"sender":		sender,
					"recipient":	recipient,
					"cc":			self.Field("cc"),
					"bcc":			self.Field("bcc"),
					"subject":		self.Field("subject"),
					"body":			self.Field("body"),
					"attachment":	self.Field("attachment", as_string=False),
				}










from os													import path as ospath
from pygwarts.magical.philosophers_stone				import Transmutable
from pygwarts.magical.philosophers_stone.transmutations	import ControlledTransmutation
from pygwarts.magical.spells							import geminio
from pygwarts.magical.spells							import flagrate
from pygwarts.magical.spells							import patronus








def byte_size_string(value :int | float | str) -> str | None :

	"""
		Utility function that accepts a numeric "value" and calculates the size in bytes which that value
		represents to return it as pretty string. The "value" must be convertible to float and will be
		finally converted to integer. The result string will consist of numbers and letters according
		to relevant bytes. The sign will be preserved, allowing to obtain strings with negative size.
		Maximum size is 999 terabyte, according to modern (2024) technologies. If "value" conversion
		failed or if "value" is more than 15 digits, None will be returned.
	"""

	try:

		byte_size = str(int(float(value)))
		sign = "-" if byte_size[0] == "-" else ""
		byte_size = byte_size.lstrip("-")

	except:	return
	else:
		if	len(byte_size) <16:
			return	sign + " ".join(

				f"{int(S)}{B}" for S,B in zip(

					( byte_size.zfill(15)[i:i+3] for i in range(0,16,3) ),
					( "T", "G", "M", "K", "B", )

				)	if int(S)
			)		or "0B"








class TextWrapper(ControlledTransmutation):

	"""
		pygwarts utility decorator, that serves as an instrument for text editing, which is just addition
		of a text before (header) and after (footer). As ControlledTransmutation class, accepts
		corresponding "header" and "footer" arguments, which must be strings and defaulted to empty strings,
		in case only one is needed. In mutable chain acts as a mutation - takes decorated Transmutable,
		which upon call must return a string value to be modified by the decoration. Returns final string,
		or None if only decorated object returns not a string. Doesn't handles any Exceptions, only
		explicitly converts "header" and "footer" to strings at initiation time and once again in
		decoration's interpolation.
	"""

	def __init__(self, header :str =str(), footer :str =str()):

		self.header = str(header)
		self.footer = str(footer)

	def _mutable_chain_injection(self, layer :Transmutable) -> Transmutable :

		header = self.header
		footer = self.footer


		class TextWrapping(geminio(layer)):
			def __call__(self, *args, **kwargs) -> str | None :

				if	isinstance(current_value := super().__call__(*args, **kwargs), str):

					final_value = f"{header}{current_value}{footer}"
					vlen = len(final_value)
					self.loggy.debug(f"Wrapped text now {vlen} symbol{flagrate(vlen)}")


					return	final_value


				self.loggy.debug("Text value to wrap not found")
		return	TextWrapping








class WriteWrapper(ControlledTransmutation):

	"""
		pygwarts utility decorator, that serves as an instrument for saving text to a file. As
		ControlledTransmutation class, accepts "path" as string where to write text, and two boolean
		flags, "rewrite" which decides whether "path" will be opened in "w" or "a" mode, and "text_mode"
		which decides whether written text will be returned or a "path" in case of success. In mutable chain
		acts as a mutation - takes decorated Transmutable, which upon call must return a text to be written.
		Returns None if neither text could be obtained nor an exception was raised in writing time.
	"""

	def __init__(self, path :str, *, rewrite :bool =False, text_mode :bool =True):

		self.path = path
		self.rewrite = rewrite
		self.text_mode = text_mode

	def _mutable_chain_injection(self, layer :Transmutable) -> Transmutable :

		path = self.path
		rewrite = self.rewrite
		text_mode = self.text_mode


		class WriteWrapping(geminio(layer)):
			def __call__(self, *args, **kwargs) -> str | None :


				if	isinstance(current_text := super().__call__(*args, **kwargs), str):
					self.loggy.debug(f"Writing {(tl := len(current_text))} symbol{flagrate(tl)}")


					if	isinstance(path, str):

							try:

								with open(path, "w" if rewrite else "a") as descriptor:

									descriptor.write(current_text)
									self.loggy.info(f"Written to {path}")

							except	Exception as E : self.loggy.debug(f"Writing failed due to {patronus(E)}")
							else:	return current_text if text_mode else path


					else:	self.loggy.debug("Path to write is not a string")
				else:		self.loggy.debug("Text to write not fetched")


		return	WriteWrapping










from __future__							import annotations
from datetime							import datetime
from datetime							import timedelta
from typing								import Dict
from typing								import Tuple
from collections.abc					import Sequence
from math								import copysign
from re									import Pattern
from pygwarts.magical.time_turner.utils	import VALID_DATE_1_P
from pygwarts.magical.time_turner.utils	import VALID_DATE_2_P
from pygwarts.magical.time_turner.utils	import VALID_DATE_3_P
from pygwarts.magical.time_turner.utils	import VALID_DATE_4_P
from pygwarts.magical.time_turner.utils	import VALID_TIME_P
from pygwarts.magical.time_turner.utils	import monthmap








class TimeTurner:

	r"""
		                                           ..::::-----:::..                                           
		                                  .:-=+*##%%%%%%%%%%%%%%%%%%##*+==-:.                                 
		                            .:-+*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%#+=-...                          
		                       ..:=*%%%%%%%%%%%#*+==-:::......:::--=++*#%%%%%%%%%%#+-..                       
		                     .=*%%%%%%%%#*=-:..             . ..       ..:-=+#%%%%%%%%#=:..                   
		                 ..=*%%%%%%%*+-..      .:-==++***####****+==-:...    ..:=*%%%%%%%#+:.                 
		               .-*%%%%%%#+-..   ..-=+*%%%%%%%%%%%%%%%%%%%%%%%%%%#*=-:..   .:=#%%%%%%#=.               
		             .=#%%%%%%+:. .  .-+#%%%%%%%%#*+==-+%%%%%%#-==+*##%%%%%%%#*=:.    .=#%%%%%%+:             
		           .=#%%%%%#=..  ..=*%%%%%%#+=-....    -#%%%%#*     ...:-+*%%%%%%#+:.   .:+%%%%%%*:.          
		         .-#%%%%%*:.. ..-*%%%%%#+-.             .*%%%.             .:=*%%%%%#+:.   .=%%%%%%+.         
		        .*%%%%%#:   ..=%%%%%#=:.         .:-=++**%%%%**++=-:..        ..-*%%%%%*:.   .+%%%%%#-        
		      .-%%%%%%=.   .+%%%%%*:.       .:=*#%#%%%%%*=::-+#%#%%%%%*+-.        .=#%%%%*:    :#%%%%%+..     
		     .=%%%%%*.   .=%%%%%+..     ..=*%%%%%%%%#+:.       :=*%%%%*%%%#+:.      .-#%%%%*..  .=%%%%%*.     
		    .+%%%%%+.   :#%%%%*.     ..-*%%%*-*%%#+:.            ..=**=*#%%%%#=..     .=%%%%%-.  .-%%%%%#..   
		   .+%%%%%=.   :%%%%%-      .-#%%%%%%*%*-.                  .:+%%%%#%%%%+.     .:#%%%%+.  .:%%%%%#.   
		  .=%%%%%+.   -%%%%#:.    .:#%#*%%%%%#-.                       .*%%%%%%%%%=.     .+%%%%+.  .:%%%%%*.  
		  :%%%%%*.   :%%%%#.     .=%%%%%%%%%%:..                    .....*%%%%#*%%%*..    .=%%%%+.  .-%%%%%+. 
		 .*%%%%%.  ..#%%%#.     .+%%%%+%%%%%# :--::..            ...::-- =%%%#=:*#%%#.     .+%%%%-.   +%%%%%: 
		 :%%%%%=.  .+%%%%-     .=%%%*+:+*%%%%-.:-----:::.......:::-----..#%%%%%#%%%%%#.     .#%%%%.   .%%%%%+ 
		.=%%%%%..  .%%%%#     ..%%%*+=.=+*%%%%+..:-------------------:.-#%%%%%%%%%%%%%=      =%%%%-    *%%%%#.
		*%%%%%%+. .-%%%%=     .+%%%%%%=%%%%%%%%%*-..::-----------::.:+#%%%%%%%%%%%#+%%%.     :%%%%*..:*#%%%%%*
		%%%%%%%%####%%%%-      #%%%%%%%%%%%%#*#%%%%#+-:..----:.:-=*%%%%%%#*-*%%%%%%%%%%:     .%%%%%###%%%%%%%%
		%%%%%%%%%%%%%%%%-.    .#%%#%%%%%%%%#+-*#%%%%%#*-.:---..*#%%%%%%%*+-:=*#%%%%%%%%:     .%%%%%%%%%%%%%%%%
		%%%%%%%#---+%%%%-.    .*%%%%%%%%%%%%%%%%%#+-.    .--:    .:=*%%%%%%+%%%%%%%#%%%.     .%%%%#--=%%%%%%%%
		:*%%%%%:   :%%%%+.     -%%%%##%%%%%%%%%+:.       .--.       ..-#%%%%%%%%%%%%%%*.     -%%%%+. .:*%%%%%:
		.-%%%%%:.  .#%%%%.     .#%%#=+#%%%%%%*.         ..::....       .=%%%%%%%%#%%%%:      *%%%%:   .#%%%%*.
		 .%%%%%*.  .-%%%%+.    .:#%%%%%%%#%%%......::::-------::::....  .+%%%%%#=:*%%=      :%%%%*.   -%%%%%- 
		  =%%%%%-    +%%%%=.     :#%%%%%%%%%%..:---------------------::. =%%%%%%%*%%=      .#%%%#.   .#%%%%#. 
		  .#%%%%%:.  .*%%%%=.     .*%%%%%%%%%+.:----------------------:.:%%%%%%%%%#:.    ..#%%%%:  ..*%%%%%:. 
		  ..#%%%%#..  .*%%%%+.     .-#%%%#+-*#*:.:------------------:..+%%%%%%%%%+.     .-#%%%%:.  .+%%%%%=.  
		    :%%%%%#:   .+%%%%#:.    ..-#%%#+#%%%#=..:------------::.-+%%*-*#%%%+.     ..+%%%%#.   .+%%%%%=    
		     :#%%%%#:.   -#%%%%+.      .:+#%%%%%%%%*-..:------::.:+#%%%%%#%%*-.      .-#%%%%+.   .*%%%%%=.    
		      .#%%%%%+.  ..+%%%%%+.       .:=#%%%%%%%%*=:.::..-+%%#%%%%%#+-. .     .=#%%%%*:    -#%%%%%-      
		       .+%%%%%#-.   .+%%%%%*-.      . .:=+*#%%%%%#++*%%%%%#*+=-..        :+%%%%%*-.   .+%%%%%#:..     
		        .:#%%%%%*:.  ..=#%%%%%+-.            .:::#%%%-::..           .:+#%%%%%*:.   .+%%%%%%=.        
		          .=%%%%%%*-.   .:+%%%%%%*+-:.         :-#%%%=-          .:=*%%%%%%*-..  .:+%%%%%%+.          
		           ..=%%%%%%#=.    .-+#%%%%%%%*+=-:... =%%%%%%# ...::-+*#%%%%%%%*-.    .-*%%%%%%*:.           
		              .-#%%%%%%*=..   ..-+*%%%%%%%%%%###%%%%%%%##%%%%%%%%%%#+=:.   ..-+%%%%%%%+:.             
		                .:+%%%%%%%#+-...  ...-=+*##%%%%%%%%%%%%%%%%%#*+=-:..    .:=*%%%%%%%*-..               
		                  ..-+%%%%%%%%#+-:.      ....:::------:::.....    ...-=*%%%%%%%%*=.                   
		                      .:+*%%%%%%%%%#+=-::...              ....:-=+*#%%%%%%%%#+-..                     
		                         ..-=*#%%%%%%%%%%%%##***+++++++***#%%%%%%%%%%%%%*+-..                         
		                             . .:=+*#%%%%%%%%%%%%%%%%%%%%%%%%%%%%#*+=-:..                             
		                                     ..:-==++***######***++==-::.. .                                  
		                                                          ..       
		The datetime library functionality wrapper. Takes full date in format:
			D{1,2}[,/.-\s:\](M{1,2}|NAME)[,/.-\s:\]YYYY
			(M{1,2}|NAME)[,/.-\s:\]D{1,2}[,/.-\s:\]YYYY
			YYYY[,/.-\s:\](M{1,2}|NAME)[,/.-\s:\]D{1,2}
			YYYY[,/.-\s:\]D{1,2}[,/.-\s:\](M{1,2}|NAME)
		and time in format:
			M{1,2}[,/.-\s:\]?H{1,2}[,/.-\s:\]?(S?{1,2})?
		The date sequence will be considered in four different variations, and it is up to user not to
		mess with it. The default TimeTurner sequence is DD/MM/YYYY. It is used in TimeTurner logic and
		it is recommended to adhere such format. Current implementation allow to use only full year format,
		that implies ONLY FOUR DIGITS YEAR IS ACCEPTABLE! Time sequence is considered in only one way, but
		is possible to provided in any ways, such as with or without lead digits, with or without delimiters
		and with or without seconds (hours and minutes are mandatory). Anyway, the default format defined
		by TimeTurner is HH:MM:SS and it is highly recommended to adhere such format, and to not mess with
		time sequence at all, cause different formats may be misinterpreted by the user at first place.
		Both date and time sequences are provided by "datepoint" and "timepoint" arguments. Both can be type
		string, tuple of strings/integers, float/int timestamp, dateimte or even TimeTurner objects. Both
		can be omitted, and in such case datepoint defaulted to current moment (by datetime.today call),
		timepoint defaulted to midnight. It is important to notice, that if datepoint provided as datetime
		or TimeTurner (even when defaulted to current moment by datetime.today call), the timepoint will be
		first obtained from datepoint, and then, if provided as argument, will be tuned to that value.
		There are properties that are default datetime format codes implementations and datetime
		functionality access. Any other datetime library functionality might be accessed via POINT attribute.
	"""

	REPR_MODES = {

		"aspath":	"/",
		"aswpath":	"\\",
		"asjoin":	"",
		"ascolon":	":",
		"dashed":	"-",
		"spaced":	" ",
	}




	# The following properties are default datetime format codes implementations,
	# that the 1989 C standard requires, and which work on all platforms
	# with a standard C implementation. It's almost all directives, except some
	# ISO and locale's representations.
	#
	# %a - Weekday as locale’s abbreviated name (Sun, Mon, ...);
	# %A - Weekday as locale’s full name (Sunday, Monday, …);
	# %w - Weekday as a decimal number, where 0 is Sunday and 6 is Saturday;
	# %d - Day of the month as a zero-padded decimal number;
	# %b - Month as locale’s abbreviated name (Jan, Feb, …);
	# %B - Month as locale’s full name (January, February, …);
	# %m - Month as a zero-padded decimal number;
	# %y - Year without century as a zero-padded decimal number;
	# %Y - Year with century as a decimal number;
	# %H - Hour (24-hour clock) as a zero-padded decimal number;
	# %I - Hour (12-hour clock) as a zero-padded decimal number;
	# %p - Locale’s equivalent of either AM or PM (it's literally AM or PM);
	# %M - Minute as a zero-padded decimal number;
	# %S - Second as a zero-padded decimal number;
	# %f - Microsecond as a decimal number, zero-padded to 6 digits;
	# %z - UTC offset in the form ±HHMM[SS[.ffffff]] (empty string if the object is naive);
	# %Z - Time zone name (empty string if the object is naive);
	# %j - Day of the year as a zero-padded decimal number;
	# %U - Week number of the year as a zero-padded decimal number (week start from Sunday);
	# %W - Week number of the year as a zero-padded decimal number (week start from Monday);
	@property
	def a(self) -> str : return self.POINT.strftime("%a")
	@property
	def A(self) -> str : return self.POINT.strftime("%A")
	@property
	def w(self) -> str : return self.POINT.strftime("%w")
	@property
	def d(self) -> str : return self.POINT.strftime("%d")
	@property
	def b(self) -> str : return self.POINT.strftime("%b")
	@property
	def B(self) -> str : return self.POINT.strftime("%B")
	@property
	def m(self) -> str : return self.POINT.strftime("%m")
	@property
	def y(self) -> str : return self.POINT.strftime("%y")
	@property
	def Y(self) -> str : return self.POINT.strftime("%Y")
	@property
	def H(self) -> str : return self.POINT.strftime("%H")
	@property
	def I(self) -> str : return self.POINT.strftime("%I")
	@property
	def p(self) -> str : return self.POINT.strftime("%p")
	@property
	def M(self) -> str : return self.POINT.strftime("%M")
	@property
	def S(self) -> str : return self.POINT.strftime("%S")
	@property
	def f(self) -> str : return self.POINT.strftime("%f")
	@property
	def z(self) -> str : return self.POINT.strftime("%z")
	@property
	def Z(self) -> str : return self.POINT.strftime("%Z")
	@property
	def j(self) -> str : return self.POINT.strftime("%j")
	@property
	def U(self) -> str : return self.POINT.strftime("%U")
	@property
	def W(self) -> str : return self.POINT.strftime("%W")


	# One can use this property that redirect to datetime strftime method,
	# to obtain desired special formatted strings.
	@property
	def format(self) -> datetime : return self.POINT.strftime


	# Local date corresponding to the POSIX timestamp
	@property
	def epoch(self) -> float : return self.POINT.timestamp()


	# Boolean that decide if current datetime object corresponds to
	# the very first day of the month.
	@property
	def is_first_day(self) -> bool : return self.POINT.strftime("%d") == "01"


	# Boolean that decide if current year is a leap year.
	@property
	def is_leap_year(self) -> bool :

		year = int(self.POINT.strftime("%Y"))
		return bool(not year%4 and (year%100 or not year%400))




	def travel(self, **traveling :Dict[str,int]) -> TimeTurner :

		"""
			Wrapper over timedelta to provide months delta.
			By months delta only leap to last date of previous month or first day of next month is meant.
		"""

		# Casting values to integer is the argument guard, cause timedelta itself takes arguments as
		# integers and all logic below will not account for months being float, so if desired month will
		# be float and infinity loop will occur due to integer current_month never will equal float
		# desired_month. This infinity loop will be stopped by OverflowError as 2914029 subdays
		# doesn't fit for timedelta. However, such case and other are guarded by this casts.
		_travelings = { k: int(v) for k,v in traveling.items() }


		try:

			m = _travelings["months"]
			del _travelings["months"]


			start = self.POINT


			# As month value not restricted and can be as large as possible, desired
			# month is obtained by modulo of start point month plus number of months
			# to travel. If desired month is divided by 12 it's defaulted to 12.
			current_month = int(start.strftime("%m"))
			desired_month = (int(current_month) + m)%12 or 12


			# Subdays and months_passed will be incremented in loop by difference,
			# which is solved as one day step with the sign, obtained from number
			# of months to travel, as a direction (minus for backward and plus fo forward).
			# Traveling months is the loop first stopping condition.
			subdays = 0
			months_passed = 0
			traveling_months = abs(m)
			diff = int(copysign(m**0, m))


			# By travel provided months is meant the obtaining as many days, as need to
			# to obtain such a desired month. On each iteration timedelta will be
			# summed with star point to produce subdays. When the desired month
			# will be reached, the amount of subdays will produce the last day of
			# desired month, in case of backwards traveling, or the first dat of month,
			# in case of forward traveling. This days will be passed along with the rest
			# of traveling keywords to the final timedelta constructor.
			while traveling_months != months_passed or desired_month != current_month:

				subdays += diff
				current_point = start + timedelta(days=subdays)
				new_month = int(current_point.strftime("%m"))


				if	new_month != current_month:

					months_passed += 1
					current_month = new_month


			if	"days" in _travelings:

				_travelings["days"] += subdays
			else:
				_travelings["days"] = subdays


		# KeyError means there were no months to convert, so simple timedelta object created,
		# with provided traveling. Any invalid traveling keyword argument passed will raise
		# timedelta TypeError. In case of timedelta success construction, current object POINT
		# will be modified with this timedelta, so travel happened.
		except	KeyError : pass
		self.POINT += timedelta(**_travelings)


		return	self




	def sight(self, **sightings :Dict[str,int]) -> TimeTurner :

		"""
			Duplicate current TimeTurner object and travels "sightings" keyword arguments.
			Returns new TimeTurner object, which is traveled from current.
		"""

		return	TimeTurner(self).travel(**sightings)




	def diff(	self,
				subtrahend			:TimeTurner | datetime | float | int | str	=None,
			)-> float | TypeError	:

		"""
			Method to obtain difference between current TimeTurner object and any number-like value,
			which might be derived from another TimerTurner or datetime object. The returned value is
			float seconds between current TimeTurner object timestamp and "subtrahend", with sign
			that point to the direction:
				>0 - current TimeTurner object has older timestamp
				<0 - current TimeTurner object has younger timestamp
			Raises TypeError in case of invalid "subtrahend" value.
		"""

		match subtrahend:

			case None				: return self.POINT.timestamp() - datetime.today().timestamp()
			case TimeTurner()		: return self.POINT.timestamp() - subtrahend.POINT.timestamp()
			case datetime()			: return self.POINT.timestamp() - subtrahend.timestamp()
			case float() | int()	: return self.POINT.timestamp() - subtrahend
			case str()				:

				try:	return	self.POINT.timestamp() - float(subtrahend)
				except	ValueError	: pass
			case _					: pass


		raise	TypeError(f"Can't calculate difference between \"{subtrahend}\" and TimeTurner object")




	def __str__(self) : return self.POINT.strftime("%A %d/%m/%Y %H:%M:%S")
	def __add__(self, addition :float | int | str) -> TimeTurner | TypeError :

		"""
			Addition to current TimeTurner object of seconds as any number-interpretable value,
			which will be double cast to allow floatable string to provide seconds.
			Returned value is the new TimeTurner object, simply sighted by provided "addition"
			cast value. Raises exception in case of invalid "addition" operand.
		"""

		try:	return	self.sight(seconds=int(float(addition)))
		except:	raise	TypeError(f"Value \"{addition}\" cannot be added to TimeTurner object")




	def __sub__(self, subtraction :float | int | str) -> TimeTurner | TypeError :

		"""
			Subtraction from current TimeTurner object of seconds as any number-interpretable value,
			which will be double cast to allow floatable string to provide seconds.
			Returned value is the new TimeTurner object, simply sighted by provided "subtraction"
			cast value. Raises exception in case of invalid "subtraction" operand.
		"""

		try:	return	self.sight(seconds=-int(float(subtraction)))
		except:	raise	TypeError(f"Value \"{subtraction}\" cannot be subtracted from TimeTurner object")




	def __eq__(self, other :TimeTurner | datetime | float | int | str) -> bool :

		"""
			Comparison of equality of current TimeTurner object timestamp with any number-like value,
			which might be derived from another TimerTurner or datetime object. Returns False
			in any cases, when "other" operand cannot be interpreted as number. In terms of stability,
			values comparison occurs after int cast.
		"""

		match other:

			case TimeTurner()		: return int(self.POINT.timestamp()) == int(other.epoch)
			case datetime()			: return int(self.POINT.timestamp()) == int(other.timestamp())
			case float() | int()	: return int(self.POINT.timestamp()) == int(other)
			case str()				:

				try		: return	int(self.POINT.timestamp()) == int(float(other))
				except	: pass
			case _		: pass


		return	False




	def __gt__(self, other :TimeTurner | datetime | float | int | str) -> bool | TypeError :

		"""
			Comparison of greatness of current TimeTurner object timestamp with any number-like value,
			which might be derived from another TimerTurner or datetime object. Raises exception in case
			of invalid "other" operand. In terms of stability, values comparison occurs after int cast.
		"""

		match other:

			case TimeTurner()		: return int(other.epoch)		<int(self.POINT.timestamp())
			case datetime()			: return int(other.timestamp())	<int(self.POINT.timestamp())
			case float() | int()	: return int(other)				<int(self.POINT.timestamp())
			case str()				:

				try		: return	int(float(other)) <int(self.POINT.timestamp())
				except	: pass
			case _		: pass


		raise	TypeError(f"Object \"{other}\" cannot be compared with TimeTurner object")




	def __ge__(self, other :TimeTurner | datetime | float | int | str) -> bool | TypeError :

		"""
			Comparison of greatness or equality of current TimeTurner object timestamp with any
			number-like value, which might be derived from another TimerTurner or datetime object.
			Raises exception in case of invalid "other" operand. In terms of stability,
			values comparison occurs after int cast.
		"""

		match other:

			case TimeTurner()		: return int(other.epoch)		<= int(self.POINT.timestamp())
			case datetime()			: return int(other.timestamp())	<= int(self.POINT.timestamp())
			case float() | int()	: return int(other)				<= int(self.POINT.timestamp())
			case str()				:

				try		: return	int(float(other)) <= int(self.POINT.timestamp())
				except	: pass
			case _		: pass


		raise	TypeError(f"Object \"{other}\" cannot be compared with TimeTurner object")




	def __lt__(self, other :TimeTurner | datetime | float | int | str) -> bool | TypeError :

		"""
			Comparison of lesserness of current TimeTurner object timestamp with any number-like value,
			which might be derived from another TimerTurner or datetime object. Raises exception in case
			of invalid "other" operand. In terms of stability, values comparison occurs after int cast.
		"""

		match other:

			case TimeTurner()		: return int(self.POINT.timestamp()) <int(other.epoch)
			case datetime()			: return int(self.POINT.timestamp()) <int(other.timestamp())
			case float() | int()	: return int(self.POINT.timestamp()) <int(other)
			case str()				:

				try		: return	int(self.POINT.timestamp()) <int(float(other))
				except	: pass
			case _		: pass


		raise	TypeError(f"Object \"{other}\" cannot be compared with TimeTurner object")




	def __le__(self, other :TimeTurner | datetime | float | int | str) -> bool | TypeError :

		"""
			Comparison of lesserness or equality of current TimeTurner object timestamp with any
			number-like value, which might be derived from another TimerTurner or datetime object.
			Raises exception in case of invalid "other" operand. In terms of stability,
			values comparison occurs after int cast.
		"""

		match other:

			case TimeTurner()		: return int(self.POINT.timestamp()) <= int(other.epoch)
			case datetime()			: return int(self.POINT.timestamp()) <= int(other.timestamp())
			case float() | int()	: return int(self.POINT.timestamp()) <= int(other)
			case str()				:

				try		: return	int(self.POINT.timestamp()) <= int(float(other))
				except	: pass
			case _		: pass


		raise	TypeError(f"Object \"{other}\" cannot be compared with TimeTurner object")




	def __getattr__(self, attr :str):

		"""
			Accessing attribute by parsing directives, which must be default datetime format codes,
			and join it by the representation mode (REPR_MODES key). The directives must be separated
			with mode by underscore.

			The following means TimeTurner not supposed to have attributes that starts
			with just underscore, so if such an attribute wanna be accessed, there'll be
			some raise.	Also when some non existent attribute wanna be accessed and it will lead
			to this dunder,	ValueError will be raised by extracting from split. All this raises
			will be redirected to AttributeError with special message.

			Implemented representation modes:
			_aspath		- joined by /
			_aswpath	- joined by \
			_asjoin		- joined by ""
			_ascolon	- joined by :
			_dashed		- joined by -
			_spaced		- joined by " "
		"""

		try:

			directives, mode	= attr.split("_")
			if not directives	: raise ValueError


			# For default datetime formats, that are provided by properties, getattr is used
			# to invoke corresponding property. This is used instead of strftime cause, despite
			# strftime from POINT is the proper way to handle inner logic, strftime with some
			# implicitly defined for datetime properties may break TimeTurner interface and
			# lead to undefined behavior. So to resume, if object implements some interface,
			# such interface must be used, especially if such interface is the interface to
			# some another interface.
			return	self.REPR_MODES[mode].join( getattr(self, d) for d in directives )
		except		(ValueError, KeyError):
			raise	AttributeError(f"Couldn't solve \"{attr}\" attribute")




	def __init__(
					self,
					datepoint	:str | int | float | Sequence[str|int|float] | TimeTurner | datetime =None,
					timepoint	:str | int | float | Sequence[str|int|float] | TimeTurner | datetime =None,
					**travels	:Dict[str,int],
				):

		# If datepoint contain invalid value, there'll be ValueError raise.
		# If datepoint not provided, current moment will be taken. It is important, that in this part
		# timepoint will be taken from datepoint, if it possible. Provided timepoint argument will
		# rewrite it in next part, so current TimeTurner object time will be tuned.
		match datepoint:

			case	str() | [
					str() | int(),
					str() | int(),
					str() | int()
						]			: self.POINT = datetime(*map(int, self.validate(datepoint)))
			case	int() | float()	: self.POINT = datetime.fromtimestamp(datepoint)
			case	TimeTurner()	: self.POINT = datepoint.POINT
			case	datetime()		: self.POINT = datepoint
			case	None			: self.POINT = datetime.today()
			case	_				: raise ValueError(f"Invalid datepoint \"{datepoint}\" initiation")


		# If timepoint contain invalid value, there'll be ValueError raise.
		# If timepoint not provided, current TimeTurner object time taken from datepoint.
		match timepoint:

			case	str() | [
					str() | int(), str() | int() ] | [
					str() | int(), str() | int(), str() | int() ] :

				point = self.valitime(timepoint)
				self.POINT = self.POINT.replace(

					hour=int(point[0]),
					minute=int(point[1]),
					second=int(point[2])
				)

			case	int() | float() :

				# Resolving UTC format for zero timepoint, which means timepoint initiation
				# implies strict midnight point, with no timezone reference.
				if	timepoint == 0 or timepoint == 0. :

					point = datetime.utcfromtimestamp(timepoint)
				else:
					point = datetime.fromtimestamp(timepoint)

				self.POINT = self.POINT.replace(

					hour=int(point.strftime("%H")),
					minute=int(point.strftime("%M")),
					second=int(point.strftime("%S")),
					microsecond=int(point.strftime("%f"))
				)

			case	TimeTurner() :

				self.POINT = self.POINT.replace(

					hour=int(timepoint.H),
					minute=int(timepoint.M),
					second=int(timepoint.S),
					microsecond=int(timepoint.f)
				)

			case	datetime() :

				self.POINT = self.POINT.replace(

					hour=int(timepoint.strftime("%H")),
					minute=int(timepoint.strftime("%M")),
					second=int(timepoint.strftime("%S")),
					microsecond=int(timepoint.strftime("%f"))
				)

			case	None	: pass
			case	_		: raise ValueError(f"Invalid timepoint \"{timepoint}\" initiation")


		# Final step is to maintain some traveling, if provided, and saving initiated points
		# for some further use, e.g. for sight. Points saved in form, that is default for TimeTurner.
		if	travels : self.travel(**travels)


		self.starting_datepoint = self.POINT.strftime("%d/%m/%Y")
		self.starting_timepoint = self.POINT.strftime("%H:%M:%S")




	def dsolve(self, point :str | Sequence[str]) -> Pattern | ValueError :

		"""
			Parsing out date and time using 4 different patterns:
				DD MM YYYY
				MM DD YYYY
				YYYY MM DD
				YYYY DD MM
			Returns a match if any, or raises ValueError.
		"""

		if	not isinstance(point, str):

			# Assuming any Exception at this point are cause of "point" is not iterable, or
			# iterable that containing not strings, so AttributeError must be raised.
			try:	_point = "/".join(map(str, point))
			except:	raise AttributeError(f"datepoint must be type of string or contain strings")
		else :		_point = point


		match1 = VALID_DATE_1_P.fullmatch(_point)
		if match1 : return match1


		match2 = VALID_DATE_2_P.fullmatch(_point)
		if match2 : return match2


		match3 = VALID_DATE_3_P.fullmatch(_point)
		if match3 : return match3


		match4 = VALID_DATE_4_P.fullmatch(_point)
		if match4 : return match4


		# None of 4 patterns produces result so datepoint is invalid
		raise ValueError(f"\"{_point}\" cannot be solved as valid date")




	def validate(self, point :str | Sequence[str]) -> Tuple[str,str,str] | ValueError :

		"""
			Validate date as a real date value, even accounts for leap year. Returns a tuple of
			three numeric strings, that represents year, month, day, in case of successful "point"
			solving. Raises ValueError otherwise.
		"""

		tnd, allM, thd, thdM, tod, todM, Y = self.dsolve(point).group(
			"tnd", "allM", "thd", "thdM", "tod", "todM", "y"
		)


		if	tnd == "29" and allM == "02":
			year = int(Y)

			if	not not year%4 and (year%100 or not year%400):
				raise ValueError(f"{year} is not a leap year")


		return	Y, monthmap(todM or thdM or allM), tod or thd or tnd




	def valitime(self, point :str | Sequence[str]) -> Tuple[str,str,str] | ValueError :

		"""
			Validate time as a real time value. Will try to make a string if it is not string but iterable.
			This is combined with validation cause it's much simpler then for date. Returns a tuple of
			three numeric zero-padded strings, that represents hours, minutes and seconds, in case of
			successful "point" solving. Raises ValueError otherwise.
		"""

		if	point and not isinstance(point, str):

			# Assuming any Exception at this point are cause of "point" is not iterable, or
			# iterable that containing not strings, so AttributeError must be raised.
			try:	_point = ":".join(map(lambda E : str(E).zfill(2), point))
			except:	raise AttributeError(f"timepoint must be type of string or contain strings")
		else:		_point = point
		try:

			hh,h, mm,m, ss,s = VALID_TIME_P.fullmatch(_point).group("hh","h", "mm","m", "ss","s")
			H,M,S = hh or h, mm or m, ss or s


			return	H,M, S or "00"
		except		AttributeError:
			raise	ValueError(f"Unable to validate \"{_point}\" time")










import	os
import	re
import	unittest
from	pygwarts.magical.philosophers_stone		import Transmutable
from	pygwarts.tests.irma						import IrmaTestCase
from	pygwarts.irma.contrib					import LibraryContrib
from	pygwarts.irma.access.handlers			import AccessHandler
from	pygwarts.irma.access.handlers			import AccessHandlerCounter
from	pygwarts.irma.access.handlers			import AccessHandlerRegisterCounter
from	pygwarts.irma.access.handlers.counters	import AccessCounter
from	pygwarts.irma.access.handlers.parsers	import GroupParser
from	pygwarts.irma.access.handlers.parsers	import TargetHandler
from	pygwarts.irma.access.handlers.parsers	import TargetNumberAccumulator
from	pygwarts.irma.access.handlers.parsers	import TargetStringAccumulator
from	pygwarts.irma.access.bookmarks			import VolumeBookmark
from	pygwarts.irma.access.volume				import LibraryVolume








class AccessHandlerCase(IrmaTestCase):

	"""
		Testing for fields and edge cases
	"""

	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:
			if	os.path.isfile(cls.ACCESS_HANDLER): os.remove(cls.ACCESS_HANDLER)

	@classmethod
	def setUpClass(cls): cls.make_loggy_file(cls, cls.ACCESS_HANDLER)
	def test_no_upper_init(self):

		class LonelyHandler(AccessHandler):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_HANDLER
				init_name	= "no_upper_init"
				init_level	= 10

		with self.assertLogs("no_upper_init", 10) as case_loggy : self.test_case = LonelyHandler()
		self.assertIn(

			f"INFO:no_upper_init:AccessHandler {self.test_case} not assigned to any bookmark",
			case_loggy.output
		)


	def test_bookmark_init(self):
		class LonelyBookmark(VolumeBookmark):

			class LonelyHandler(AccessHandler):	pass
			class loggy(LibraryContrib):

				handler		= self.ACCESS_HANDLER
				init_name	= "bookmark_init"
				init_level	= 10

		with self.assertLogs("bookmark_init", 10) as case_loggy : self.test_case = LonelyBookmark()
		self.assertIn(

			f"INFO:bookmark_init:Bookmark {self.test_case} not assigned to library",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:bookmark_init:Assigned to {self.test_case} bookmark",
			case_loggy.output
		)


	def test_no_bookmark_init(self):
		class NotBookmark(Transmutable):

			class NotBookmarkHandler(AccessHandler):	pass
			class loggy(LibraryContrib):

				handler		= self.ACCESS_HANDLER
				init_name	= "no_bookmark_init"
				init_level	= 10

		with self.assertLogs("no_bookmark_init", 10) as case_loggy : self.test_case = NotBookmark()
		self.assertIn(

			f"INFO:no_bookmark_init:AccessHandler {self.test_case.NotBookmarkHandler} "
			"not assigned to any bookmark",
			case_loggy.output
		)








	def test_AccessHandlerRegisterCounter_no_link(self):

		class TestVolume(LibraryVolume):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_HANDLER
				init_name	= "AccessHandlerRegisterCounter_no_link"
				init_level	= 10

			class TestBookmark(VolumeBookmark):

				trigger = "place for trigger"
				@AccessHandlerRegisterCounter
				class Handler(AccessHandler):

					# Imitating handler that makes no link
					def __call__(self, line :str, volume :LibraryVolume): return True

		with self.assertLogs("AccessHandlerRegisterCounter_no_link", 10) as case_loggy:

			self.test_case = TestVolume()
			self.test_case.TestBookmark.Handler("this is line", self.test_case)

		self.assertIsNone(self.test_case[self.test_case.TestBookmark.Handler])
		self.assertIn(

			"DEBUG:AccessHandlerRegisterCounter_no_link:"
			f"{self.test_case} counter link was not created",
			case_loggy.output
		)








	def test_AccessCounter(self):

		class TestVolume(LibraryVolume):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_HANDLER
				init_name	= "AccessCounter"
				init_level	= 10

			class TestBookmark(VolumeBookmark):

				trigger = "place for trigger"
				class Counter(AccessCounter):	pass

		with self.assertLogs("AccessCounter", 10) as case_loggy:

			self.test_case = TestVolume()
			self.test_case.TestBookmark.Counter("line", self.test_case)
			self.test_case.TestBookmark.Counter("line", self.test_case)
			self.test_case.TestBookmark.Counter("line", self.test_case)

		self.assertIn(
			f"INFO:AccessCounter:Volume {self.test_case} not assigned to any library", case_loggy.output
		)
		self.assertIn(f"DEBUG:AccessCounter:Assigned to volume {self.test_case}", case_loggy.output)
		self.assertIn(
			f"DEBUG:AccessCounter:Assigned to {self.test_case.TestBookmark} bookmark", case_loggy.output
		)

		self.assertIn("DEBUG:AccessCounter:Handler counter incremented to 1", case_loggy.output)
		self.assertIn("DEBUG:AccessCounter:Handler counter incremented to 2", case_loggy.output)
		self.assertIn("DEBUG:AccessCounter:Handler counter incremented to 3", case_loggy.output)
		self.assertTrue(hasattr(self.test_case.TestBookmark.Counter, "access_handler_counter"))
		self.assertEqual(self.test_case.TestBookmark.Counter.access_handler_counter, 3)

		self.assertIn(f"DEBUG:AccessCounter:{self.test_case} counter incremented to 1", case_loggy.output)
		self.assertIn(f"DEBUG:AccessCounter:{self.test_case} counter incremented to 2", case_loggy.output)
		self.assertIn(f"DEBUG:AccessCounter:{self.test_case} counter incremented to 3", case_loggy.output)
		self.assertIn(self.test_case.TestBookmark.Counter, self.test_case)
		self.assertIn("counter", self.test_case[self.test_case.TestBookmark.Counter])
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter]["counter"], 3)








	def test_GroupParser_no_rpattern(self):

		class TestVolume(LibraryVolume):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_HANDLER
				init_name	= "GroupParser_no_rpattern"
				init_level	= 10

			class TestBookmark(VolumeBookmark):

				trigger = "place for trigger"
				class Counter(GroupParser):	pass

		with self.assertLogs("GroupParser_no_rpattern", 10) as case_loggy : self.test_case = TestVolume()
		self.assertIn(

			f"WARNING:GroupParser_no_rpattern:{self.test_case.TestBookmark.Counter} "
			"doesn't have pattern for parsing",
			case_loggy.output
		)




	def test_GroupParser_type_rpattern(self):

		class TestVolume(LibraryVolume):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_HANDLER
				init_name	= "GroupParser_type_rpattern"
				init_level	= 10

			class TestBookmark(VolumeBookmark):

				trigger = "place for trigger"
				class Counter(GroupParser): rpattern = 42

		with self.assertLogs("GroupParser_type_rpattern", 10) as case_loggy : self.test_case = TestVolume()
		self.assertIn(

			f"WARNING:GroupParser_type_rpattern:{self.test_case.TestBookmark.Counter} "
			"has invalid pattern for parsing",
			case_loggy.output
		)








	def test_TargetHandler_str(self):

		class TestVolume(LibraryVolume):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_HANDLER
				init_name	= "TargetHandler_str"
				init_level	= 10

			class TestBookmark(VolumeBookmark):

				trigger = "place for trigger"
				@AccessHandlerCounter
				@AccessHandlerRegisterCounter
				class Counter(TargetHandler): rpattern = r"(?P<target>42)"

		with self.assertLogs("TargetHandler_str", 10) as case_loggy:

			self.test_case = TestVolume()
			self.test_case.TestBookmark.Counter("the answer is 42", self.test_case)

		self.assertIn(
			"DEBUG:TargetHandler_str:Accepted string \"(?P<target>42)\"", case_loggy.output
		)
		self.assertIn("DEBUG:TargetHandler_str:Recap set to \"42\"", case_loggy.output)
		self.assertIsInstance(self.test_case[self.test_case.TestBookmark.Counter], dict)
		self.assertIn("recap", self.test_case[self.test_case.TestBookmark.Counter])
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter].get("recap"), "42")
		self.assertIn("DEBUG:TargetHandler_str:Handler counter incremented to 1", case_loggy.output)
		self.assertEqual(self.test_case.TestBookmark.Counter.access_handler_counter, 1)
		self.assertIn(
			f"DEBUG:TargetHandler_str:{self.test_case} counter incremented to 1", case_loggy.output
		)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter]["counter"], 1)


		self.test_case[self.test_case.TestBookmark.Counter]["recap"] = "69"
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter].get("recap"), "69")


		with self.assertLogs("TargetHandler_str", 10) as case_loggy:
			self.test_case.TestBookmark.update("the answer is 42", self.test_case)

		self.assertIn("DEBUG:TargetHandler_str:Recap set to \"42\"", case_loggy.output)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter].get("recap"), "42")
		self.assertIn("DEBUG:TargetHandler_str:Handler counter incremented to 2", case_loggy.output)
		self.assertEqual(self.test_case.TestBookmark.Counter.access_handler_counter, 2)
		self.assertIn(
			f"DEBUG:TargetHandler_str:{self.test_case} counter incremented to 2", case_loggy.output
		)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter]["counter"], 2)


		with self.assertLogs("TargetHandler_str", 10) as case_loggy:
			self.test_case.TestBookmark.update("Better be 69", self.test_case)

		self.assertIn("DEBUG:TargetHandler_str:No match for target", case_loggy.output)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter].get("recap"), "42")
		self.assertIn("DEBUG:TargetHandler_str:Handler counter incremented to 3", case_loggy.output)
		self.assertEqual(self.test_case.TestBookmark.Counter.access_handler_counter, 3)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter]["counter"], 2)


		with self.assertLogs("TargetHandler_str", 10) as case_loggy:
			self.test_case.TestBookmark.Counter("Better be 69", self.test_case)

		self.assertIn("DEBUG:TargetHandler_str:No match for target", case_loggy.output)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter].get("recap"), "42")
		self.assertIn("DEBUG:TargetHandler_str:Handler counter incremented to 4", case_loggy.output)
		self.assertEqual(self.test_case.TestBookmark.Counter.access_handler_counter, 4)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter]["counter"], 2)




	def test_TargetHandler_Pattern(self):

		class TestVolume(LibraryVolume):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_HANDLER
				init_name	= "TargetHandler_Pattern"
				init_level	= 10

			class TestBookmark(VolumeBookmark):

				trigger = "place for trigger"
				@AccessHandlerCounter
				@AccessHandlerRegisterCounter
				class Counter(TargetHandler): rpattern = re.compile(r"(?P<target>42)")

		with self.assertLogs("TargetHandler_Pattern", 10) as case_loggy:

			self.test_case = TestVolume()
			self.test_case.TestBookmark.Counter("the answer is 42", self.test_case)

		self.assertIn(
			"DEBUG:TargetHandler_Pattern:Accepted pattern \"(?P<target>42)\"", case_loggy.output
		)
		self.assertIn("DEBUG:TargetHandler_Pattern:Recap set to \"42\"", case_loggy.output)
		self.assertIsInstance(self.test_case[self.test_case.TestBookmark.Counter], dict)
		self.assertIn("recap", self.test_case[self.test_case.TestBookmark.Counter])
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter].get("recap"), "42")
		self.assertIn("DEBUG:TargetHandler_Pattern:Handler counter incremented to 1", case_loggy.output)
		self.assertEqual(self.test_case.TestBookmark.Counter.access_handler_counter, 1)
		self.assertIn(
			f"DEBUG:TargetHandler_Pattern:{self.test_case} counter incremented to 1", case_loggy.output
		)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter]["counter"], 1)


		self.test_case[self.test_case.TestBookmark.Counter]["recap"] = "69"
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter].get("recap"), "69")


		with self.assertLogs("TargetHandler_Pattern", 10) as case_loggy:
			self.test_case.TestBookmark.update("the answer is 42", self.test_case)

		self.assertIn("DEBUG:TargetHandler_Pattern:Recap set to \"42\"", case_loggy.output)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter].get("recap"), "42")
		self.assertIn("DEBUG:TargetHandler_Pattern:Handler counter incremented to 2", case_loggy.output)
		self.assertEqual(self.test_case.TestBookmark.Counter.access_handler_counter, 2)
		self.assertIn(
			f"DEBUG:TargetHandler_Pattern:{self.test_case} counter incremented to 2", case_loggy.output
		)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter]["counter"], 2)


		with self.assertLogs("TargetHandler_Pattern", 10) as case_loggy:
			self.test_case.TestBookmark.update("Better be 69", self.test_case)

		self.assertIn("DEBUG:TargetHandler_Pattern:No match for target", case_loggy.output)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter].get("recap"), "42")
		self.assertIn("DEBUG:TargetHandler_Pattern:Handler counter incremented to 3", case_loggy.output)
		self.assertEqual(self.test_case.TestBookmark.Counter.access_handler_counter, 3)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter]["counter"], 2)


		with self.assertLogs("TargetHandler_Pattern", 10) as case_loggy:
			self.test_case.TestBookmark.Counter("Better be 69", self.test_case)

		self.assertIn("DEBUG:TargetHandler_Pattern:No match for target", case_loggy.output)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter].get("recap"), "42")
		self.assertIn("DEBUG:TargetHandler_Pattern:Handler counter incremented to 4", case_loggy.output)
		self.assertEqual(self.test_case.TestBookmark.Counter.access_handler_counter, 4)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter]["counter"], 2)








	def test_TargetNumberAccumulator_str(self):

		class TestVolume(LibraryVolume):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_HANDLER
				init_name	= "TargetNumberAccumulator_str"
				init_level	= 10

			class TestBookmark(VolumeBookmark):

				trigger = "place for trigger"
				@AccessHandlerCounter
				@AccessHandlerRegisterCounter
				class Counter(TargetNumberAccumulator): rpattern = r"(?P<target>42)"

		with self.assertLogs("TargetNumberAccumulator_str", 10) as case_loggy:

			self.test_case = TestVolume()
			self.test_case.TestBookmark.Counter("the answer is 42", self.test_case)

		self.assertIn(
			"DEBUG:TargetNumberAccumulator_str:Accepted string \"(?P<target>42)\"", case_loggy.output
		)
		self.assertIn("DEBUG:TargetNumberAccumulator_str:Recap accumulation is \"42\"", case_loggy.output)
		self.assertIsInstance(self.test_case[self.test_case.TestBookmark.Counter], dict)
		self.assertIn("recap", self.test_case[self.test_case.TestBookmark.Counter])
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter].get("recap"), 42)
		self.assertIn(
			"DEBUG:TargetNumberAccumulator_str:Handler counter incremented to 1", case_loggy.output
		)
		self.assertEqual(self.test_case.TestBookmark.Counter.access_handler_counter, 1)
		self.assertIn(

			f"DEBUG:TargetNumberAccumulator_str:{self.test_case} counter incremented to 1",
			case_loggy.output
		)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter]["counter"], 1)


		with self.assertLogs("TargetNumberAccumulator_str", 10) as case_loggy:
			self.test_case.TestBookmark.update("the answer is 42", self.test_case)

		self.assertIn("DEBUG:TargetNumberAccumulator_str:Recap accumulation is \"84\"", case_loggy.output)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter].get("recap"), 84)
		self.assertIn(
			"DEBUG:TargetNumberAccumulator_str:Handler counter incremented to 2", case_loggy.output
		)
		self.assertEqual(self.test_case.TestBookmark.Counter.access_handler_counter, 2)
		self.assertIn(

			f"DEBUG:TargetNumberAccumulator_str:{self.test_case} counter incremented to 2",
			case_loggy.output
		)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter]["counter"], 2)


		with self.assertLogs("TargetNumberAccumulator_str", 10) as case_loggy:
			self.test_case.TestBookmark.update("Better be 69", self.test_case)

		self.assertIn("DEBUG:TargetNumberAccumulator_str:No match for target", case_loggy.output)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter].get("recap"), 84)
		self.assertIn(
			"DEBUG:TargetNumberAccumulator_str:Handler counter incremented to 3", case_loggy.output
		)
		self.assertEqual(self.test_case.TestBookmark.Counter.access_handler_counter, 3)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter]["counter"], 2)


		with self.assertLogs("TargetNumberAccumulator_str", 10) as case_loggy:
			self.test_case.TestBookmark.Counter("Better be 69", self.test_case)

		self.assertIn("DEBUG:TargetNumberAccumulator_str:No match for target", case_loggy.output)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter].get("recap"), 84)
		self.assertIn(
			"DEBUG:TargetNumberAccumulator_str:Handler counter incremented to 4", case_loggy.output
		)
		self.assertEqual(self.test_case.TestBookmark.Counter.access_handler_counter, 4)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter]["counter"], 2)




	def test_TargetNumberAccumulator_Pattern(self):

		class TestVolume(LibraryVolume):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_HANDLER
				init_name	= "TargetNumberAccumulator_Pattern"
				init_level	= 10

			class TestBookmark(VolumeBookmark):

				trigger = "place for trigger"
				@AccessHandlerCounter
				@AccessHandlerRegisterCounter
				class Counter(TargetNumberAccumulator): rpattern = re.compile(r"(?P<target>42)")

		with self.assertLogs("TargetNumberAccumulator_Pattern", 10) as case_loggy:

			self.test_case = TestVolume()
			self.test_case.TestBookmark.Counter("the answer is 42", self.test_case)

		self.assertIn(
			"DEBUG:TargetNumberAccumulator_Pattern:Accepted pattern \"(?P<target>42)\"", case_loggy.output
		)
		self.assertIn("DEBUG:TargetNumberAccumulator_Pattern:Recap accumulation is \"42\"", case_loggy.output)
		self.assertIsInstance(self.test_case[self.test_case.TestBookmark.Counter], dict)
		self.assertIn("recap", self.test_case[self.test_case.TestBookmark.Counter])
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter].get("recap"), 42)
		self.assertIn(
			"DEBUG:TargetNumberAccumulator_Pattern:Handler counter incremented to 1", case_loggy.output
		)
		self.assertEqual(self.test_case.TestBookmark.Counter.access_handler_counter, 1)
		self.assertIn(

			f"DEBUG:TargetNumberAccumulator_Pattern:{self.test_case} counter incremented to 1",
			case_loggy.output
		)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter]["counter"], 1)


		with self.assertLogs("TargetNumberAccumulator_Pattern", 10) as case_loggy:
			self.test_case.TestBookmark.update("the answer is 42", self.test_case)

		self.assertIn("DEBUG:TargetNumberAccumulator_Pattern:Recap accumulation is \"84\"", case_loggy.output)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter].get("recap"), 84)
		self.assertIn(
			"DEBUG:TargetNumberAccumulator_Pattern:Handler counter incremented to 2", case_loggy.output
		)
		self.assertEqual(self.test_case.TestBookmark.Counter.access_handler_counter, 2)
		self.assertIn(

			f"DEBUG:TargetNumberAccumulator_Pattern:{self.test_case} counter incremented to 2",
			case_loggy.output
		)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter]["counter"], 2)


		with self.assertLogs("TargetNumberAccumulator_Pattern", 10) as case_loggy:
			self.test_case.TestBookmark.update("Better be 69", self.test_case)

		self.assertIn("DEBUG:TargetNumberAccumulator_Pattern:No match for target", case_loggy.output)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter].get("recap"), 84)
		self.assertIn(
			"DEBUG:TargetNumberAccumulator_Pattern:Handler counter incremented to 3", case_loggy.output
		)
		self.assertEqual(self.test_case.TestBookmark.Counter.access_handler_counter, 3)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter]["counter"], 2)


		with self.assertLogs("TargetNumberAccumulator_Pattern", 10) as case_loggy:
			self.test_case.TestBookmark.Counter("Better be 69", self.test_case)

		self.assertIn("DEBUG:TargetNumberAccumulator_Pattern:No match for target", case_loggy.output)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter].get("recap"), 84)
		self.assertIn(
			"DEBUG:TargetNumberAccumulator_Pattern:Handler counter incremented to 4", case_loggy.output
		)
		self.assertEqual(self.test_case.TestBookmark.Counter.access_handler_counter, 4)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter]["counter"], 2)








	def test_TargetStringAccumulator_str(self):

		class TestVolume(LibraryVolume):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_HANDLER
				init_name	= "TargetStringAccumulator_str"
				init_level	= 10

			class TestBookmark(VolumeBookmark):

				trigger = "place for trigger"
				@AccessHandlerCounter
				@AccessHandlerRegisterCounter
				class Counter(TargetStringAccumulator): rpattern = r"(?P<target>42)"

		with self.assertLogs("TargetStringAccumulator_str", 10) as case_loggy:

			self.test_case = TestVolume()
			self.test_case.TestBookmark.Counter("the answer is 42", self.test_case)

		self.assertIn(
			"DEBUG:TargetStringAccumulator_str:Accepted string \"(?P<target>42)\"", case_loggy.output
		)
		self.assertIn("DEBUG:TargetStringAccumulator_str:Recap extended to 1 item", case_loggy.output)
		self.assertIsInstance(self.test_case[self.test_case.TestBookmark.Counter], dict)
		self.assertIn("recap", self.test_case[self.test_case.TestBookmark.Counter])
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter].get("recap"),[ "42" ])
		self.assertIn(
			"DEBUG:TargetStringAccumulator_str:Handler counter incremented to 1", case_loggy.output
		)
		self.assertEqual(self.test_case.TestBookmark.Counter.access_handler_counter, 1)
		self.assertIn(

			f"DEBUG:TargetStringAccumulator_str:{self.test_case} counter incremented to 1",
			case_loggy.output
		)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter]["counter"], 1)


		with self.assertLogs("TargetStringAccumulator_str", 10) as case_loggy:
			self.test_case.TestBookmark.update("the answer is 42", self.test_case)

		self.assertIn("DEBUG:TargetStringAccumulator_str:Recap extended to 2 items", case_loggy.output)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter].get("recap"),[ "42","42" ])
		self.assertIn(
			"DEBUG:TargetStringAccumulator_str:Handler counter incremented to 2", case_loggy.output
		)
		self.assertEqual(self.test_case.TestBookmark.Counter.access_handler_counter, 2)
		self.assertIn(

			f"DEBUG:TargetStringAccumulator_str:{self.test_case} counter incremented to 2",
			case_loggy.output
		)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter]["counter"], 2)


		with self.assertLogs("TargetStringAccumulator_str", 10) as case_loggy:
			self.test_case.TestBookmark.update("Better be 69", self.test_case)

		self.assertIn("DEBUG:TargetStringAccumulator_str:No match for target", case_loggy.output)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter].get("recap"),[ "42","42" ])
		self.assertIn(
			"DEBUG:TargetStringAccumulator_str:Handler counter incremented to 3", case_loggy.output
		)
		self.assertEqual(self.test_case.TestBookmark.Counter.access_handler_counter, 3)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter]["counter"], 2)


		with self.assertLogs("TargetStringAccumulator_str", 10) as case_loggy:
			self.test_case.TestBookmark.Counter("Better be 69", self.test_case)

		self.assertIn("DEBUG:TargetStringAccumulator_str:No match for target", case_loggy.output)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter].get("recap"),[ "42","42" ])
		self.assertIn(
			"DEBUG:TargetStringAccumulator_str:Handler counter incremented to 4", case_loggy.output
		)
		self.assertEqual(self.test_case.TestBookmark.Counter.access_handler_counter, 4)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter]["counter"], 2)




	def test_TargetStringAccumulator_Pattern(self):

		class TestVolume(LibraryVolume):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_HANDLER
				init_name	= "TargetStringAccumulator_Pattern"
				init_level	= 10

			class TestBookmark(VolumeBookmark):

				trigger = "place for trigger"
				@AccessHandlerCounter
				@AccessHandlerRegisterCounter
				class Counter(TargetStringAccumulator): rpattern = re.compile(r"(?P<target>42)")

		with self.assertLogs("TargetStringAccumulator_Pattern", 10) as case_loggy:

			self.test_case = TestVolume()
			self.test_case.TestBookmark.Counter("the answer is 42", self.test_case)

		self.assertIn(
			"DEBUG:TargetStringAccumulator_Pattern:Accepted pattern \"(?P<target>42)\"", case_loggy.output
		)
		self.assertIn("DEBUG:TargetStringAccumulator_Pattern:Recap extended to 1 item", case_loggy.output)
		self.assertIsInstance(self.test_case[self.test_case.TestBookmark.Counter], dict)
		self.assertIn("recap", self.test_case[self.test_case.TestBookmark.Counter])
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter].get("recap"),[ "42" ])
		self.assertIn(
			"DEBUG:TargetStringAccumulator_Pattern:Handler counter incremented to 1", case_loggy.output
		)
		self.assertEqual(self.test_case.TestBookmark.Counter.access_handler_counter, 1)
		self.assertIn(

			f"DEBUG:TargetStringAccumulator_Pattern:{self.test_case} counter incremented to 1",
			case_loggy.output
		)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter]["counter"], 1)


		with self.assertLogs("TargetStringAccumulator_Pattern", 10) as case_loggy:
			self.test_case.TestBookmark.update("the answer is 42", self.test_case)

		self.assertIn("DEBUG:TargetStringAccumulator_Pattern:Recap extended to 2 items", case_loggy.output)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter].get("recap"),[ "42","42" ])
		self.assertIn(
			"DEBUG:TargetStringAccumulator_Pattern:Handler counter incremented to 2", case_loggy.output
		)
		self.assertEqual(self.test_case.TestBookmark.Counter.access_handler_counter, 2)
		self.assertIn(

			f"DEBUG:TargetStringAccumulator_Pattern:{self.test_case} counter incremented to 2",
			case_loggy.output
		)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter]["counter"], 2)


		with self.assertLogs("TargetStringAccumulator_Pattern", 10) as case_loggy:
			self.test_case.TestBookmark.update("Better be 69", self.test_case)

		self.assertIn("DEBUG:TargetStringAccumulator_Pattern:No match for target", case_loggy.output)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter].get("recap"),[ "42","42" ])
		self.assertIn(
			"DEBUG:TargetStringAccumulator_Pattern:Handler counter incremented to 3", case_loggy.output
		)
		self.assertEqual(self.test_case.TestBookmark.Counter.access_handler_counter, 3)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter]["counter"], 2)


		with self.assertLogs("TargetStringAccumulator_Pattern", 10) as case_loggy:
			self.test_case.TestBookmark.Counter("Better be 69", self.test_case)

		self.assertIn("DEBUG:TargetStringAccumulator_Pattern:No match for target", case_loggy.output)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter].get("recap"),[ "42","42" ])
		self.assertIn(
			"DEBUG:TargetStringAccumulator_Pattern:Handler counter incremented to 4", case_loggy.output
		)
		self.assertEqual(self.test_case.TestBookmark.Counter.access_handler_counter, 4)
		self.assertEqual(self.test_case[self.test_case.TestBookmark.Counter]["counter"], 2)








if	__name__ == "__main__" : unittest.main(verbosity=2)










import	os
import	unittest
from	pathlib						import Path
from	shutil						import rmtree
from	pygwarts.tests.hagrid		import HagridTestCase
from	pygwarts.irma.contrib		import LibraryContrib
from	pygwarts.hagrid.thrivables	import Tree
from	pygwarts.hagrid.thrivables	import Copse
from	pygwarts.hagrid.planting	import Flourish
from	pygwarts.hagrid.bloom.twigs	import Germination
from	pygwarts.hagrid.bloom.leafs	import Rejuvenation








class DispatchingCases(HagridTestCase):

	"""
		Some cases for Flourish algorithm.
	"""

	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:
			if	os.path.isfile(cls.FLOURISH_HANDLER): os.remove(cls.FLOURISH_HANDLER)

		if	os.path.isdir(cls.INIT_SET_BOUGH_1): rmtree(cls.INIT_SET_BOUGH_1)
		if	os.path.isdir(cls.INIT_SET_BOUGH_2): rmtree(cls.INIT_SET_BOUGH_2)
		if	os.path.isdir(cls.INIT_SET_BOUGH_3): rmtree(cls.INIT_SET_BOUGH_3)

	@classmethod
	def setUpClass(cls):

		if	not os.path.isdir(cls.INIT_SET_BOUGH_1): os.makedirs(cls.INIT_SET_BOUGH_1)
		if	not os.path.isdir(cls.INIT_SET_BOUGH_2): os.makedirs(cls.INIT_SET_BOUGH_2)
		if	not os.path.isdir(cls.INIT_SET_BOUGH_3): os.makedirs(cls.INIT_SET_BOUGH_3)
		cls.make_loggy_file(cls, cls.FLOURISH_HANDLER)




	def test_failed_start(self):
		class sync(Flourish):
			class loggy(LibraryContrib):

				handler		= self.FLOURISH_HANDLER
				init_name	= "failed_start"
				init_level	= 10


		with self.assertLogs("failed_start", 10) as case_loggy:

			self.test_case = sync()
			self.test_case(
				(
					self.INIT_SET_SPROUT,
					(
						(
							Path(self.INIT_SET_SPROUT),
							[ Path(self.INIT_SET_SPROUT).joinpath("some folder") ],
							[ Path(self.INIT_SET_SPROUT).joinpath("some file") ],
						)	for _ in range(1)
					)
				)
			)

		self.assertIn(f"CRITICAL:failed_start:{self.test_case} failed to start flourish", case_loggy.output)








	def test_no_plants(self):
		class Sakura(Tree):

			bough	= self.INIT_SET_BOUGH_1
			class loggy(LibraryContrib):

				handler		= self.FLOURISH_HANDLER
				init_name	= "no_plants"
				init_level	= 10

			class sync(Flourish):		pass


		self.test_case = Sakura()


		with self.assertLogs("no_plants", 10) as case_loggy:
			self.test_case.sync()

		self.assertIn("DEBUG:no_plants:No plants to flourish", case_loggy.output)








	def test_valid_generator_1(self):
		class Sakura(Tree):

			bough	= self.INIT_SET_BOUGH_1
			class loggy(LibraryContrib):

				handler		= self.FLOURISH_HANDLER
				init_name	= "valid_generator_1"
				init_level	= 10

			class sync(Flourish):		pass


		with self.assertLogs("valid_generator_1", 10) as case_loggy:

			self.test_case = Sakura()
			self.test_case.sync(

				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				[ Path(self.INIT_SET_SPROUT).joinpath("some folder") ],
				[ Path(self.INIT_SET_SPROUT).joinpath("some file") ],
			)


		self.assertIn(f"DEBUG:valid_generator_1:Flourishing {self.INIT_SET_SPROUT}", case_loggy.output)
		self.assertIn(f"DEBUG:valid_generator_1:Current branch \"{self.INIT_SET_SPROUT}\"", case_loggy.output)
		self.assertIn("DEBUG:valid_generator_1:Number of twigs: 1", case_loggy.output)
		self.assertIn("DEBUG:valid_generator_1:Number of leafs: 1", case_loggy.output)
		self.assertIn("DEBUG:valid_generator_1:Chest is empty", case_loggy.output)








	def test_valid_generator_2(self):
		class Sakura(Tree):

			bough	= self.INIT_SET_BOUGH_1
			class loggy(LibraryContrib):

				handler		= self.FLOURISH_HANDLER
				init_name	= "valid_generator_2"
				init_level	= 10

			class sync(Flourish):		pass


		with self.assertLogs("valid_generator_2", 10) as case_loggy:

			self.test_case = Sakura()
			self.test_case.sync.innervate(
				self.test_case,
				self.INIT_SET_SPROUT,
				Path(self.INIT_SET_SPROUT),
				[ Path(self.INIT_SET_SPROUT).joinpath("some folder") ],
				[ Path(self.INIT_SET_SPROUT).joinpath("some file") ],
			)


		self.assertNotIn(f"DEBUG:valid_generator_2:Flourishing {self.INIT_SET_SPROUT}", case_loggy.output)
		self.assertIn(f"DEBUG:valid_generator_2:Current branch \"{self.INIT_SET_SPROUT}\"", case_loggy.output)
		self.assertIn("DEBUG:valid_generator_2:Number of twigs: 1", case_loggy.output)
		self.assertIn("DEBUG:valid_generator_2:Number of leafs: 1", case_loggy.output)
		self.assertIn("DEBUG:valid_generator_2:Chest is empty", case_loggy.output)








	def test_invalid_generator_1(self):
		class Sakura(Tree):

			bough	= self.INIT_SET_BOUGH_1
			class loggy(LibraryContrib):

				handler		= self.FLOURISH_HANDLER
				init_name	= "invalid_generator_1"
				init_level	= 10

			class sync(Flourish):		pass


		self.test_case = Sakura()
		with self.assertLogs("invalid_generator_1", 10) as case_loggy:

			self.test_case.sync(1)

		self.assertIn(f"WARNING:invalid_generator_1:Invalid plant length 1", case_loggy.output)
		with self.assertLogs("invalid_generator_1", 10) as case_loggy:

			self.test_case.sync(1,2)

		self.assertIn(f"WARNING:invalid_generator_1:Invalid plant length 2", case_loggy.output)
		with self.assertLogs("invalid_generator_1", 10) as case_loggy:

			self.test_case.sync(1,2,3)

		self.assertIn(f"WARNING:invalid_generator_1:Invalid plant length 3", case_loggy.output)
		with self.assertLogs("invalid_generator_1", 10) as case_loggy:

			self.test_case.sync(1,2,3,4,5)

		self.assertIn(f"WARNING:invalid_generator_1:Invalid plant length 5", case_loggy.output)








	def test_invalid_generator_2(self):
		class Sakura(Tree):

			bough	= self.INIT_SET_BOUGH_1
			class loggy(LibraryContrib):

				handler		= self.FLOURISH_HANDLER
				init_name	= "invalid_generator_2"
				init_level	= 10

			class sync(Flourish):		pass


		self.test_case = Sakura()
		for S in ( 1,.1,( "tree", ),[ "tree" ],{ "tree" },{ "tree": "tree" },self.test_case, None,True ):
			with self.subTest(sprout=S):
				with self.assertLogs("invalid_generator_2", 10) as case_loggy:

					self.test_case.sync(S,2,3,4)
				self.assertIn(f"WARNING:invalid_generator_2:Invalid plant sprout \"{S}\"", case_loggy.output)








	def test_invalid_generator_3(self):
		class Sakura(Tree):

			bough	= self.INIT_SET_BOUGH_1
			class loggy(LibraryContrib):

				handler		= self.FLOURISH_HANDLER
				init_name	= "invalid_generator_3"
				init_level	= 10

			class sync(Flourish):		pass


		self.test_case = Sakura()
		for B in ( 1,.1,( "tree", ),[ "tree" ],{ "tree" },{ "tree": "tree" },self.test_case, None,True ):
			with self.subTest(branch=B):
				with self.assertLogs("invalid_generator_3", 10) as case_loggy:

					self.test_case.sync(self.INIT_SET_SPROUT,B,3,4)
				self.assertIn(f"WARNING:invalid_generator_3:Invalid plant branch \"{B}\"", case_loggy.output)








	def test_invalid_generator_4(self):
		class Sakura(Tree):

			bough	= self.INIT_SET_BOUGH_1
			class loggy(LibraryContrib):

				handler		= self.FLOURISH_HANDLER
				init_name	= "invalid_generator_4"
				init_level	= 10

			class sync(Flourish):		pass


		self.test_case = Sakura()
		for T in ( 1,.1,( "tree", ),{ "tree" },{ "tree": "tree" },self.test_case, None,True ):
			with self.subTest(twigs=T):
				with self.assertLogs("invalid_generator_4", 10) as case_loggy:

					self.test_case.sync(self.INIT_SET_SPROUT,Path(self.INIT_SET_SPROUT),T,4)
				self.assertIn(f"WARNING:invalid_generator_4:Invalid plant twigs {type(T)}", case_loggy.output)








	def test_invalid_generator_5(self):
		class Sakura(Tree):

			bough	= self.INIT_SET_BOUGH_1
			class loggy(LibraryContrib):

				handler		= self.FLOURISH_HANDLER
				init_name	= "invalid_generator_5"
				init_level	= 10

			class sync(Flourish):		pass


		self.test_case = Sakura()
		for L in ( 1,.1,( "tree", ),{ "tree" },{ "tree": "tree" },self.test_case, None,True ):
			with self.subTest(twigs=L):
				with self.assertLogs("invalid_generator_5", 10) as case_loggy:

					self.test_case.sync(self.INIT_SET_SPROUT,Path(self.INIT_SET_SPROUT),[],L)
				self.assertIn(f"WARNING:invalid_generator_5:Invalid plant leafs {type(L)}", case_loggy.output)








	def test_valid_grove(self):
		class Forest(Copse):

			class loggy(LibraryContrib):

				handler		= self.FLOURISH_HANDLER
				init_name	= "valid_grove"
				init_level	= 10

			class Deeper(Copse):
				class DeeperFirst(Tree):		bough = self.INIT_SET_BOUGH_1
				class DeeperSecond(Tree):		bough = self.INIT_SET_BOUGH_2
				class DeeperThird(Tree):		bough = self.INIT_SET_BOUGH_3

				class Deepest(Copse):
					class DeepestFirst(Tree):	bough = self.INIT_SET_BOUGH_1
					class DeepestSecond(Tree):	bough = self.INIT_SET_BOUGH_2
					class DeepestThird(Tree):	bough = self.INIT_SET_BOUGH_3

			class First(Tree):					bough = self.INIT_SET_BOUGH_1
			class Second(Tree):					bough = self.INIT_SET_BOUGH_2
			class Third(Tree):					bough = self.INIT_SET_BOUGH_3
			class sync(Flourish):				pass


		self.test_case = Forest()
		with self.assertLogs("valid_grove", 10) as case_loggy:
			forest = list(self.test_case.sync.grove(self.test_case))

		self.assertEqual(len(forest), 9)
		self.assertAlmostEqual(

			forest,
			[
				self.test_case.Deeper.DeeperFirst,
				self.test_case.Deeper.DeeperSecond,
				self.test_case.Deeper.DeeperThird,
				self.test_case.Deeper.Deepest.DeepestFirst,
				self.test_case.Deeper.Deepest.DeepestSecond,
				self.test_case.Deeper.Deepest.DeepestThird,
				self.test_case.First,
				self.test_case.Second,
				self.test_case.Third,
			]
		)
		self.assertIn(

			f"DEBUG:valid_grove:Walking copse {self.test_case} inner {self.test_case.Deeper}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:valid_grove:Walking copse {self.test_case.Deeper} inner {self.test_case.Deeper.Deepest}",
			case_loggy.output
		)








	def test_invalid_grove(self):
		class Forest(Copse):

			class loggy(LibraryContrib):

				handler		= self.FLOURISH_HANDLER
				init_name	= "invalid_grove"
				init_level	= 10

			class Deeper(Copse):
				class DeeperFirst(Tree):		bough = self.INIT_SET_BOUGH_1
				class DeeperSecond(Tree):		bough = self.INIT_SET_BOUGH_2
				class DeeperThird(Tree):		bough = self.INIT_SET_BOUGH_3

				class Deepest(Copse):
					class DeepestFirst(Tree):	bough = self.INIT_SET_BOUGH_1
					class DeepestSecond(Tree):	bough = self.INIT_SET_BOUGH_2
					class DeepestThird(Tree):	bough = self.INIT_SET_BOUGH_3

			class First(Tree):					bough = self.INIT_SET_BOUGH_1
			class Second(Tree):					bough = self.INIT_SET_BOUGH_2
			class Third(Tree):					bough = self.INIT_SET_BOUGH_3
			class sync(Flourish):				pass
			class rejuve(Rejuvenation):			pass
			class germinate(Germination):		pass


		self.test_case = Forest()
		self.test_case(self.test_case.sync)
		self.test_case.Deeper(self.test_case.rejuve)
		self.test_case.Deeper.Deepest(self.test_case.germinate)

		with self.assertLogs("invalid_grove", 10) as case_loggy:
			forest = list(self.test_case.sync.grove(self.test_case))

		self.assertEqual(len(forest), 9)
		self.assertAlmostEqual(

			forest,
			[
				self.test_case.Deeper.DeeperFirst,
				self.test_case.Deeper.DeeperSecond,
				self.test_case.Deeper.DeeperThird,
				self.test_case.Deeper.Deepest.DeepestFirst,
				self.test_case.Deeper.Deepest.DeepestSecond,
				self.test_case.Deeper.Deepest.DeepestThird,
				self.test_case.First,
				self.test_case.Second,
				self.test_case.Third,
			]
		)
		self.assertIn(

			f"DEBUG:invalid_grove:Walking copse {self.test_case} inner {self.test_case.Deeper}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:invalid_grove:Walking copse {self.test_case.Deeper} inner {self.test_case.Deeper.Deepest}",
			case_loggy.output
		)
		self.assertIn(

			f"WARNING:invalid_grove:Invalid thrivable \"{self.test_case.sync}\" for flourish",
			case_loggy.output
		)
		self.assertIn(

			f"WARNING:invalid_grove:Invalid thrivable \"{self.test_case.rejuve}\" for flourish",
			case_loggy.output
		)
		self.assertIn(

			f"WARNING:invalid_grove:Invalid thrivable \"{self.test_case.germinate}\" for flourish",
			case_loggy.output
		)








	def test_empty_grove(self):
		class Forest(Copse):

			class loggy(LibraryContrib):

				handler		= self.FLOURISH_HANDLER
				init_name	= "empty_grove"
				init_level	= 10

			class sync(Flourish):	pass


		with self.assertLogs("empty_grove", 10) as case_loggy:

			self.test_case = Forest()
			self.test_case.sync(self.INIT_SET_SPROUT, Path(self.INIT_SET_SPROUT), [],[])

		self.assertIn(f"DEBUG:empty_grove:Chest is empty", case_loggy.output)








if	__name__ == "__main__" : unittest.main(verbosity=2)







